"""Prototype for callable functions with full type safety."""
from collections.abc import Callable
from types import CellType, CodeType, MethodType
from typing import Any, overload, ParamSpec, Protocol, runtime_checkable, Self, TypeVar, TypeVarTuple
import sys

#======== Copied from typeshed:stdlib\_typeshed\__init__.pyi ========
type AnnotationForm = Any

if (3, 14) <= sys.version_info:
	from annotationlib import Format

# NOTE These return annotations, which can be arbitrary objects
	type AnnotateFunc = Callable[[Format], dict[str, AnnotationForm]]
	type EvaluateFunc = Callable[[Format], AnnotationForm]
#======== End Copied from typeshed:stdlib\_typeshed\__init__.pyi ========

@runtime_checkable
class CallableFunction[**P, R](Protocol):
	"""A Protocol representing callable functions with descriptor support.

	Mimics types.FunctionType while being a drop-in replacement for `collections.abc.Callable`. Includes all standard function
	attributes and the descriptor protocol for proper method binding.

	Note: @runtime_checkable only validates attribute presence, not signatures.
	"""

	@property
	def __closure__(self) -> tuple[CellType, ...] | None:
		"""Tuple of cells that contain bindings for the function's free variables."""
		...
	__code__: CodeType
	__defaults__: tuple[Any, ...] | None
	__dict__: dict[str, Any]
	@property
	def __globals__(self) -> dict[str, Any]:
		"""The global namespace in which the function was defined."""
		...
	__name__: str
	__qualname__: str
	__annotations__: dict[str, AnnotationForm]
	if (3, 14) <= sys.version_info:
		__annotate__: AnnotateFunc | None
	__kwdefaults__: dict[str, Any] | None
	@property
	def __builtins__(self) -> dict[str, Any]:
		"""The built-in namespace in which the function was defined."""
		...
	__type_params__: tuple[TypeVar | ParamSpec | TypeVarTuple, ...]
	__module__: str

	def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
		"""Execute the callable with the given arguments."""
		...

	@overload
	def __get__(self, instance: None, owner: type, /) -> Self: ...
	@overload
	def __get__(self, instance: object, owner: type | None = None, /) -> MethodType: ...
	def __get__(self, instance: object | None, owner: type | None = None) -> Self | MethodType:
		"""Descriptor protocol for method binding.

		Returns self when accessed on the class, or a bound MethodType when accessed on an instance.
		"""
		...



from .utils import no_default
from functools import partial, reduce
from importlib import import_module
from operator import attrgetter, not_
from types import MethodType
import inspect
import sys

PYPY = hasattr(sys, 'pypy_version_info') and sys.version_info[0] > 2
__all__ = ('identity', 'apply', 'thread_first', 'thread_last', 'memoize', 'compose', 'compose_left', 'pipe', 'complement', 'juxt', 'do', 'curry', 'flip', 'excepts')
PYPY = hasattr(sys, 'pypy_version_info')

def identity(x):
    """ Identity function. Return x

    >>> identity(3)
    3
    """
    return x

def apply(*func_and_args, **kwargs):
    """ Applies a function and returns the results

    >>> def double(x): return 2*x
    >>> def inc(x):    return x + 1
    >>> apply(double, 5)
    10

    >>> tuple(map(apply, [double, inc, double], [10, 500, 8000]))
    (20, 501, 16000)
    """
    if not func_and_args:
        raise TypeError('func argument is required')
    func, args = (func_and_args[0], func_and_args[1:])
    return func(*args, **kwargs)

def thread_first(val, *forms):
    """ Thread value through a sequence of functions/forms

    >>> def double(x): return 2*x
    >>> def inc(x):    return x + 1
    >>> thread_first(1, inc, double)
    4

    If the function expects more than one input you can specify those inputs
    in a tuple.  The value is used as the first input.

    >>> def add(x, y): return x + y
    >>> def pow(x, y): return x**y
    >>> thread_first(1, (add, 4), (pow, 2))  # pow(add(1, 4), 2)
    25

    So in general
        thread_first(x, f, (g, y, z))
    expands to
        g(f(x), y, z)

    See Also:
        thread_last
    """

    def evalform_front(val, form):
        if callable(form):
            return form(val)
        if isinstance(form, tuple):
            func, args = (form[0], form[1:])
            args = (val,) + args
            return func(*args)
    return reduce(evalform_front, forms, val)

def thread_last(val, *forms):
    """ Thread value through a sequence of functions/forms

    >>> def double(x): return 2*x
    >>> def inc(x):    return x + 1
    >>> thread_last(1, inc, double)
    4

    If the function expects more than one input you can specify those inputs
    in a tuple.  The value is used as the last input.

    >>> def add(x, y): return x + y
    >>> def pow(x, y): return x**y
    >>> thread_last(1, (add, 4), (pow, 2))  # pow(2, add(4, 1))
    32

    So in general
        thread_last(x, f, (g, y, z))
    expands to
        g(y, z, f(x))

    >>> def iseven(x):
    ...     return x % 2 == 0
    >>> list(thread_last([1, 2, 3], (map, inc), (filter, iseven)))
    [2, 4]

    See Also:
        thread_first
    """

    def evalform_back(val, form):
        if callable(form):
            return form(val)
        if isinstance(form, tuple):
            func, args = (form[0], form[1:])
            args = args + (val,)
            return func(*args)
    return reduce(evalform_back, forms, val)

def instanceproperty(fget=None, fset=None, fdel=None, doc=None, classval=None):
    """ Like @property, but returns ``classval`` when used as a class attribute

    >>> class MyClass(object):
    ...     '''The class docstring'''
    ...     @instanceproperty(classval=__doc__)
    ...     def __doc__(self):
    ...         return 'An object docstring'
    ...     @instanceproperty
    ...     def val(self):
    ...         return 42
    ...
    >>> MyClass.__doc__
    'The class docstring'
    >>> MyClass.val is None
    True
    >>> obj = MyClass()
    >>> obj.__doc__
    'An object docstring'
    >>> obj.val
    42
    """
    if fget is None:
        return partial(instanceproperty, fset=fset, fdel=fdel, doc=doc, classval=classval)
    return InstanceProperty(fget=fget, fset=fset, fdel=fdel, doc=doc, classval=classval)

class InstanceProperty(property):
    """ Like @property, but returns ``classval`` when used as a class attribute

    Should not be used directly.  Use ``instanceproperty`` instead.
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None, classval=None):
        self.classval = classval
        property.__init__(self, fget=fget, fset=fset, fdel=fdel, doc=doc)

    def __get__(self, obj, type=None):
        if obj is None:
            return self.classval
        return property.__get__(self, obj, type)

    def __reduce__(self):
        state = (self.fget, self.fset, self.fdel, self.__doc__, self.classval)
        return (InstanceProperty, state)

class curry:
    """ Curry a callable function

    Enables partial application of arguments through calling a function with an
    incomplete set of arguments.

    >>> def mul(x, y):
    ...     return x * y
    >>> mul = curry(mul)

    >>> double = mul(2)
    >>> double(10)
    20

    Also supports keyword arguments

    >>> @curry                  # Can use curry as a decorator
    ... def f(x, y, a=10):
    ...     return a * (x + y)

    >>> add = f(a=1)
    >>> add(2, 3)
    5

    See Also:
        humpy_toolz.curried - namespace of curried functions
                        https://toolz.readthedocs.io/en/latest/curry.html
    """

    def __init__(self, *args, **kwargs):
        if not args:
            raise TypeError('__init__() takes at least 2 arguments (1 given)')
        func, args = (args[0], args[1:])
        if not callable(func):
            raise TypeError('Input must be callable')
        if hasattr(func, 'func') and hasattr(func, 'args') and hasattr(func, 'keywords') and isinstance(func.args, tuple):
            _kwargs = {}
            if func.keywords:
                _kwargs.update(func.keywords)
            _kwargs.update(kwargs)
            kwargs = _kwargs
            args = func.args + args
            func = func.func
        if kwargs:
            self._partial = partial(func, *args, **kwargs)
        else:
            self._partial = partial(func, *args)
        self.__doc__ = getattr(func, '__doc__', None)
        self.__name__ = getattr(func, '__name__', '<curry>')
        self.__module__ = getattr(func, '__module__', None)
        self.__qualname__ = getattr(func, '__qualname__', None)
        self._sigspec = None
        self._has_unknown_args = None

    @instanceproperty
    def func(self):
        return self._partial.func

    @instanceproperty
    def __signature__(self):
        sig = inspect.signature(self.func)
        args = self.args or ()
        keywords = self.keywords or {}
        if is_partial_args(self.func, args, keywords, sig) is False:
            raise TypeError('curry object has incorrect arguments')
        params = list(sig.parameters.values())
        skip = 0
        for param in params[:len(args)]:
            if param.kind == param.VAR_POSITIONAL:
                break
            skip += 1
        kwonly = False
        newparams = []
        for param in params[skip:]:
            kind = param.kind
            default = param.default
            if kind == param.VAR_KEYWORD:
                pass
            elif kind == param.VAR_POSITIONAL:
                if kwonly:
                    continue
            elif param.name in keywords:
                default = keywords[param.name]
                kind = param.KEYWORD_ONLY
                kwonly = True
            else:
                if kwonly:
                    kind = param.KEYWORD_ONLY
                if default is param.empty:
                    default = no_default
            newparams.append(param.replace(default=default, kind=kind))
        return sig.replace(parameters=newparams)

    @instanceproperty
    def args(self):
        return self._partial.args

    @instanceproperty
    def keywords(self):
        return self._partial.keywords

    @instanceproperty
    def func_name(self):
        return self.__name__

    def __str__(self):
        return str(self.func)

    def __repr__(self):
        return repr(self.func)

    def __hash__(self):
        return hash((self.func, self.args, frozenset(self.keywords.items()) if self.keywords else None))

    def __eq__(self, other):
        return isinstance(other, curry) and self.func == other.func and (self.args == other.args) and (self.keywords == other.keywords)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __call__(self, *args, **kwargs):
        try:
            return self._partial(*args, **kwargs)
        except TypeError as exc:
            if self._should_curry(args, kwargs, exc):
                return self.bind(*args, **kwargs)
            raise

    def _should_curry(self, args, kwargs, exc=None):
        func = self.func
        args = self.args + args
        if self.keywords:
            kwargs = dict(self.keywords, **kwargs)
        if self._sigspec is None:
            sigspec = self._sigspec = _sigs.signature_or_spec(func)
            self._has_unknown_args = has_varargs(func, sigspec) is not False
        else:
            sigspec = self._sigspec
        if is_partial_args(func, args, kwargs, sigspec) is False:
            return False
        elif self._has_unknown_args:
            return True
        elif not is_valid_args(func, args, kwargs, sigspec):
            return True
        else:
            return False

    def bind(self, *args, **kwargs):
        return type(self)(self, *args, **kwargs)

    def call(self, *args, **kwargs):
        return self._partial(*args, **kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return curry(self, instance)

    def __reduce__(self):
        func = self.func
        modname = getattr(func, '__module__', None)
        qualname = getattr(func, '__qualname__', None)
        if qualname is None:
            qualname = getattr(func, '__name__', None)
        is_decorated = None
        if modname and qualname:
            attrs = []
            obj = import_module(modname)
            for attr in qualname.split('.'):
                if isinstance(obj, curry):
                    attrs.append('func')
                    obj = obj.func
                obj = getattr(obj, attr, None)
                if obj is None:
                    break
                attrs.append(attr)
            if isinstance(obj, curry) and obj.func is func:
                is_decorated = obj is self
                qualname = '.'.join(attrs)
                func = f'{modname}:{qualname}'
        userdict = tuple(((k, v) for k, v in self.__dict__.items() if k not in ('_partial', '_sigspec')))
        state = (type(self), func, self.args, self.keywords, userdict, is_decorated)
        return (_restore_curry, state)

def _restore_curry(cls, func, args, kwargs, userdict, is_decorated):
    if isinstance(func, str):
        modname, qualname = func.rsplit(':', 1)
        obj = import_module(modname)
        for attr in qualname.split('.'):
            obj = getattr(obj, attr)
        if is_decorated:
            return obj
        func = obj.func
    obj = cls(func, *args, **kwargs or {})
    obj.__dict__.update(userdict)
    return obj

@curry
def memoize(func, cache=None, key=None):
    """ Cache a function's result for speedy future evaluation

    Considerations:
        Trades memory for speed.
        Only use on pure functions.

    >>> def add(x, y):  return x + y
    >>> add = memoize(add)

    Or use as a decorator

    >>> @memoize
    ... def add(x, y):
    ...     return x + y

    Use the ``cache`` keyword to provide a dict-like object as an initial cache

    >>> @memoize(cache={(1, 2): 3})
    ... def add(x, y):
    ...     return x + y

    Note that the above works as a decorator because ``memoize`` is curried.

    It is also possible to provide a ``key(args, kwargs)`` function that
    calculates keys used for the cache, which receives an ``args`` tuple and
    ``kwargs`` dict as input, and must return a hashable value.  However,
    the default key function should be sufficient most of the time.

    >>> # Use key function that ignores extraneous keyword arguments
    >>> @memoize(key=lambda args, kwargs: args)
    ... def add(x, y, verbose=False):
    ...     if verbose:
    ...         print('Calculating %s + %s' % (x, y))
    ...     return x + y
    """
    if cache is None:
        cache = {}
    try:
        may_have_kwargs = has_keywords(func) is not False
        is_unary = is_arity(1, func)
    except TypeError:
        may_have_kwargs = True
        is_unary = False
    if key is None:
        if is_unary:

            def key(args, kwargs):
                return args[0]
        elif may_have_kwargs:

            def key(args, kwargs):
                return (args or None, frozenset(kwargs.items()) if kwargs else None)
        else:

            def key(args, kwargs):
                return args

    def memof(*args, **kwargs):
        k = key(args, kwargs)
        try:
            return cache[k]
        except TypeError:
            raise TypeError('Arguments to memoized function must be hashable')
        except KeyError:
            cache[k] = result = func(*args, **kwargs)
            return result
    try:
        memof.__name__ = func.__name__
    except AttributeError:
        pass
    memof.__doc__ = func.__doc__
    memof.__wrapped__ = func
    return memof

class Compose:
    """ A composition of functions

    See Also:
        compose
    """
    __slots__ = ('first', 'funcs')

    def __init__(self, funcs):
        funcs = tuple(reversed(funcs))
        self.first = funcs[0]
        self.funcs = funcs[1:]

    def __call__(self, *args, **kwargs):
        ret = self.first(*args, **kwargs)
        for f in self.funcs:
            ret = f(ret)
        return ret

    def __getstate__(self):
        return (self.first, self.funcs)

    def __setstate__(self, state):
        self.first, self.funcs = state

    @instanceproperty(classval=__doc__)
    def __doc__(self):

        def composed_doc(*fs):
            """Generate a docstring for the composition of fs.
            """
            if not fs:
                return '*args, **kwargs'
            return f'{fs[0].__name__}({composed_doc(*fs[1:])})'
        try:
            return 'lambda *args, **kwargs: ' + composed_doc(*reversed((self.first,) + self.funcs))
        except AttributeError:
            return 'A composition of functions'

    @property
    def __name__(self):
        try:
            return '_of_'.join((f.__name__ for f in reversed((self.first,) + self.funcs)))
        except AttributeError:
            return type(self).__name__

    def __repr__(self):
        return '{.__class__.__name__}{!r}'.format(self, tuple(reversed((self.first,) + self.funcs)))

    def __eq__(self, other):
        if isinstance(other, Compose):
            return other.first == self.first and other.funcs == self.funcs
        return NotImplemented

    def __ne__(self, other):
        equality = self.__eq__(other)
        return NotImplemented if equality is NotImplemented else not equality

    def __hash__(self):
        return hash(self.first) ^ hash(self.funcs)

    def __get__(self, obj, objtype=None):
        return self if obj is None else MethodType(self, obj)

    @instanceproperty
    def __signature__(self):
        base = inspect.signature(self.first)
        last = inspect.signature(self.funcs[-1])
        return base.replace(return_annotation=last.return_annotation)
    __wrapped__ = instanceproperty(attrgetter('first'))

def compose(*funcs):
    """ Compose functions to operate in series.

    Returns a function that applies other functions in sequence.

    Functions are applied from right to left so that
    ``compose(f, g, h)(x, y)`` is the same as ``f(g(h(x, y)))``.

    If no arguments are provided, the identity function (f(x) = x) is returned.

    >>> inc = lambda i: i + 1
    >>> compose(str, inc)(3)
    '4'

    See Also:
        compose_left
        pipe
    """
    if not funcs:
        return identity
    if len(funcs) == 1:
        return funcs[0]
    else:
        return Compose(funcs)

def compose_left(*funcs):
    """ Compose functions to operate in series.

    Returns a function that applies other functions in sequence.

    Functions are applied from left to right so that
    ``compose_left(f, g, h)(x, y)`` is the same as ``h(g(f(x, y)))``.

    If no arguments are provided, the identity function (f(x) = x) is returned.

    >>> inc = lambda i: i + 1
    >>> compose_left(inc, str)(3)
    '4'

    See Also:
        compose
        pipe
    """
    return compose(*reversed(funcs))

def pipe(data, *funcs):
    """ Pipe a value through a sequence of functions

    I.e. ``pipe(data, f, g, h)`` is equivalent to ``h(g(f(data)))``

    We think of the value as progressing through a pipe of several
    transformations, much like pipes in UNIX

    ``$ cat data | f | g | h``

    >>> double = lambda i: 2 * i
    >>> pipe(3, double, str)
    '6'

    See Also:
        compose
        compose_left
        thread_first
        thread_last
    """
    for func in funcs:
        data = func(data)
    return data

def complement(func):
    """ Convert a predicate function to its logical complement.

    In other words, return a function that, for inputs that normally
    yield True, yields False, and vice-versa.

    >>> def iseven(n): return n % 2 == 0
    >>> isodd = complement(iseven)
    >>> iseven(2)
    True
    >>> isodd(2)
    False
    """
    return compose(not_, func)

class juxt:
    """ Creates a function that calls several functions with the same arguments

    Takes several functions and returns a function that applies its arguments
    to each of those functions then returns a tuple of the results.

    Name comes from juxtaposition: the fact of two things being seen or placed
    close together with contrasting effect.

    >>> inc = lambda x: x + 1
    >>> double = lambda x: x * 2
    >>> juxt(inc, double)(10)
    (11, 20)
    >>> juxt([inc, double])(10)
    (11, 20)
    """
    __slots__ = ['funcs']

    def __init__(self, *funcs):
        if len(funcs) == 1 and (not callable(funcs[0])):
            funcs = funcs[0]
        self.funcs = tuple(funcs)

    def __call__(self, *args, **kwargs):
        return tuple((func(*args, **kwargs) for func in self.funcs))

    def __getstate__(self):
        return self.funcs

    def __setstate__(self, state):
        self.funcs = state

def do(func, x):
    """ Runs ``func`` on ``x``, returns ``x``

    Because the results of ``func`` are not returned, only the side
    effects of ``func`` are relevant.

    Logging functions can be made by composing ``do`` with a storage function
    like ``list.append`` or ``file.write``

    >>> from humpy_toolz import compose
    >>> from humpy_toolz.curried import do

    >>> log = []
    >>> inc = lambda x: x + 1
    >>> inc = compose(inc, do(log.append))
    >>> inc(1)
    2
    >>> inc(11)
    12
    >>> log
    [1, 11]
    """
    func(x)
    return x

@curry
def flip(func, a, b):
    """ Call the function call with the arguments flipped

    This function is curried.

    >>> def div(a, b):
    ...     return a // b
    ...
    >>> flip(div, 2, 6)
    3
    >>> div_by_two = flip(div, 2)
    >>> div_by_two(4)
    2

    This is particularly useful for built in functions and functions defined
    in C extensions that accept positional only arguments. For example:
    isinstance, issubclass.

    >>> data = [1, 'a', 'b', 2, 1.5, object(), 3]
    >>> only_ints = list(filter(flip(isinstance, int), data))
    >>> only_ints
    [1, 2, 3]
    """
    return func(b, a)

def return_none(exc):
    """ Returns None.
    """
    return None

class excepts:
    """A wrapper around a function to catch exceptions and
    dispatch to a handler.

    This is like a functional try/except block, in the same way that
    ifexprs are functional if/else blocks.

    Examples
    --------
    >>> excepting = excepts(
    ...     ValueError,
    ...     lambda a: [1, 2].index(a),
    ...     lambda _: -1,
    ... )
    >>> excepting(1)
    0
    >>> excepting(3)
    -1

    Multiple exceptions and default except clause.

    >>> excepting = excepts((IndexError, KeyError), lambda a: a[0])
    >>> excepting([])
    >>> excepting([1])
    1
    >>> excepting({})
    >>> excepting({0: 1})
    1
    """

    def __init__(self, exc, func, handler=return_none):
        self.exc = exc
        self.func = func
        self.handler = handler

    def __call__(self, *args, **kwargs):
        try:
            return self.func(*args, **kwargs)
        except self.exc as e:
            return self.handler(e)

    @instanceproperty(classval=__doc__)
    def __doc__(self):
        from textwrap import dedent
        exc = self.exc
        try:
            if isinstance(exc, tuple):
                exc_name = '(%s)' % ', '.join(map(attrgetter('__name__'), exc))
            else:
                exc_name = exc.__name__
            return dedent('                A wrapper around {inst.func.__name__!r} that will except:\n                {exc}\n                and handle any exceptions with {inst.handler.__name__!r}.\n\n                Docs for {inst.func.__name__!r}:\n                {inst.func.__doc__}\n\n                Docs for {inst.handler.__name__!r}:\n                {inst.handler.__doc__}\n                ').format(inst=self, exc=exc_name)
        except AttributeError:
            return type(self).__doc__

    @property
    def __name__(self):
        exc = self.exc
        try:
            if isinstance(exc, tuple):
                exc_name = '_or_'.join(map(attrgetter('__name__'), exc))
            else:
                exc_name = exc.__name__
            return f'{self.func.__name__}_excepting_{exc_name}'
        except AttributeError:
            return 'excepting'

def _check_sigspec(sigspec, func, builtin_func, *builtin_args):
    if sigspec is None:
        try:
            sigspec = inspect.signature(func)
        except (ValueError, TypeError) as e:
            sigspec = e
    if isinstance(sigspec, ValueError):
        return (None, builtin_func(*builtin_args))
    elif not isinstance(sigspec, inspect.Signature):
        if func in _sigs.signatures and (hasattr(func, '__signature__') and hasattr(func.__signature__, '__get__')):
            val = builtin_func(*builtin_args)
            return (None, val)
        return (None, False)
    return (sigspec, None)
if PYPY:
    _check_sigspec_orig = _check_sigspec

    def _check_sigspec(sigspec, func, builtin_func, *builtin_args):
        if func in _sigs.signatures:
            val = builtin_func(*builtin_args)
            return (None, val)
        return _check_sigspec_orig(sigspec, func, builtin_func, *builtin_args)
_check_sigspec.__doc__ = " Private function to aid in introspection compatibly across Python versions.\n\nIf a callable doesn't have a signature (Python 3) or an argspec (Python 2),\nthe signature registry in humpy_toolz._signatures is used.\n"

def num_required_args(func, sigspec=None):
    sigspec, rv = _check_sigspec(sigspec, func, _sigs._num_required_args, func)
    if sigspec is None:
        return rv
    return sum((1 for p in sigspec.parameters.values() if p.default is p.empty and p.kind in (p.POSITIONAL_OR_KEYWORD, p.POSITIONAL_ONLY)))

def has_varargs(func, sigspec=None):
    sigspec, rv = _check_sigspec(sigspec, func, _sigs._has_varargs, func)
    if sigspec is None:
        return rv
    return any((p.kind == p.VAR_POSITIONAL for p in sigspec.parameters.values()))

def has_keywords(func, sigspec=None):
    sigspec, rv = _check_sigspec(sigspec, func, _sigs._has_keywords, func)
    if sigspec is None:
        return rv
    return any((p.default is not p.empty or p.kind in (p.KEYWORD_ONLY, p.VAR_KEYWORD) for p in sigspec.parameters.values()))

def is_valid_args(func, args, kwargs, sigspec=None):
    sigspec, rv = _check_sigspec(sigspec, func, _sigs._is_valid_args, func, args, kwargs)
    if sigspec is None:
        return rv
    try:
        sigspec.bind(*args, **kwargs)
    except TypeError:
        return False
    return True

def is_partial_args(func, args, kwargs, sigspec=None):
    sigspec, rv = _check_sigspec(sigspec, func, _sigs._is_partial_args, func, args, kwargs)
    if sigspec is None:
        return rv
    try:
        sigspec.bind_partial(*args, **kwargs)
    except TypeError:
        return False
    return True

def is_arity(n, func, sigspec=None):
    """ Does a function have only n positional arguments?

    This function relies on introspection and does not call the function.
    Returns None if validity can't be determined.

    >>> def f(x):
    ...     return x
    >>> is_arity(1, f)
    True
    >>> def g(x, y=1):
    ...     return x + y
    >>> is_arity(1, g)
    False
    """
    sigspec, rv = _check_sigspec(sigspec, func, _sigs._is_arity, n, func)
    if sigspec is None:
        return rv
    num = num_required_args(func, sigspec)
    if num is not None:
        num = num == n
        if not num:
            return False
    varargs = has_varargs(func, sigspec)
    if varargs:
        return False
    keywords = has_keywords(func, sigspec)
    if keywords:
        return False
    if num is None or varargs is None or keywords is None:
        return None
    return True
num_required_args.__doc__ = " Number of required positional arguments\n\n    This function relies on introspection and does not call the function.\n    Returns None if validity can't be determined.\n\n    >>> def f(x, y, z=3):\n    ...     return x + y + z\n    >>> num_required_args(f)\n    2\n    >>> def g(*args, **kwargs):\n    ...     pass\n    >>> num_required_args(g)\n    0\n    "
has_varargs.__doc__ = " Does a function have variadic positional arguments?\n\n    This function relies on introspection and does not call the function.\n    Returns None if validity can't be determined.\n\n    >>> def f(*args):\n    ...    return args\n    >>> has_varargs(f)\n    True\n    >>> def g(**kwargs):\n    ...    return kwargs\n    >>> has_varargs(g)\n    False\n    "
has_keywords.__doc__ = " Does a function have keyword arguments?\n\n    This function relies on introspection and does not call the function.\n    Returns None if validity can't be determined.\n\n    >>> def f(x, y=0):\n    ...     return x + y\n\n    >>> has_keywords(f)\n    True\n    "
is_valid_args.__doc__ = " Is ``func(*args, **kwargs)`` a valid function call?\n\n    This function relies on introspection and does not call the function.\n    Returns None if validity can't be determined.\n\n    >>> def add(x, y):\n    ...     return x + y\n\n    >>> is_valid_args(add, (1,), {})\n    False\n    >>> is_valid_args(add, (1, 2), {})\n    True\n    >>> is_valid_args(map, (), {})\n    False\n\n    **Implementation notes**\n    Python 2 relies on ``inspect.getargspec``, which only works for\n    user-defined functions.  Python 3 uses ``inspect.signature``, which\n    works for many more types of callables.\n\n    Many builtins in the standard library are also supported.\n    "
is_partial_args.__doc__ = " Can partial(func, *args, **kwargs)(*args2, **kwargs2) be a valid call?\n\n    Returns True *only* if the call is valid or if it is possible for the\n    call to become valid by adding more positional or keyword arguments.\n\n    This function relies on introspection and does not call the function.\n    Returns None if validity can't be determined.\n\n    >>> def add(x, y):\n    ...     return x + y\n\n    >>> is_partial_args(add, (1,), {})\n    True\n    >>> is_partial_args(add, (1, 2), {})\n    True\n    >>> is_partial_args(add, (1, 2, 3), {})\n    False\n    >>> is_partial_args(map, (), {})\n    True\n\n    **Implementation notes**\n    Python 2 relies on ``inspect.getargspec``, which only works for\n    user-defined functions.  Python 3 uses ``inspect.signature``, which\n    works for many more types of callables.\n\n    Many builtins in the standard library are also supported.\n    "
from . import _signatures as _sigs




"""
Alternate namespace for cytoolz such that all functions are curried

Currying provides implicit partial evaluation of all functions

Example:

    Get usually requires two arguments, an index and a collection
    >>> from humpy_cytoolz.curried import get
    >>> get(0, ('a', 'b'))
    'a'

    When we use it in higher order functions we often want to pass a partially
    evaluated form
    >>> data = [(1, 2), (11, 22), (111, 222)]
    >>> list(map(lambda seq: get(0, seq), data))
    [1, 11, 111]

    The curried version allows simple expression of partial evaluation
    >>> list(map(get(0), data))
    [1, 11, 111]

See Also
--------
    humpy_cytoolz.functoolz.curry
"""
from . import operator
from .exceptions import merge, merge_with
from humpy_cytoolz import (
	apply, comp, complement, compose, compose_left, concat, concatv, count, curry, diff, first, flip, frequencies,
	identity, interleave, isdistinct, isiterable, juxt, last, memoize, merge_sorted, peek, pipe, second, thread_first,
	thread_last)
import humpy_cytoolz

accumulate = humpy_cytoolz.curry(humpy_cytoolz.accumulate)
assoc = humpy_cytoolz.curry(humpy_cytoolz.assoc)
assoc_in = humpy_cytoolz.curry(humpy_cytoolz.assoc_in)
cons = humpy_cytoolz.curry(humpy_cytoolz.cons)
countby = humpy_cytoolz.curry(humpy_cytoolz.countby)
dissoc = humpy_cytoolz.curry(humpy_cytoolz.dissoc)
do = humpy_cytoolz.curry(humpy_cytoolz.do)
drop = humpy_cytoolz.curry(humpy_cytoolz.drop)
excepts = humpy_cytoolz.curry(humpy_cytoolz.excepts)
filter = humpy_cytoolz.curry(humpy_cytoolz.filter)  # noqa: A001
get = humpy_cytoolz.curry(humpy_cytoolz.get)
get_in = humpy_cytoolz.curry(humpy_cytoolz.get_in)
groupby = humpy_cytoolz.curry(humpy_cytoolz.groupby)
interpose = humpy_cytoolz.curry(humpy_cytoolz.interpose)
itemfilter = humpy_cytoolz.curry(humpy_cytoolz.itemfilter)
itemmap = humpy_cytoolz.curry(humpy_cytoolz.itemmap)
iterate = humpy_cytoolz.curry(humpy_cytoolz.iterate)
join = humpy_cytoolz.curry(humpy_cytoolz.join)
keyfilter = humpy_cytoolz.curry(humpy_cytoolz.keyfilter)
keymap = humpy_cytoolz.curry(humpy_cytoolz.keymap)
map = humpy_cytoolz.curry(humpy_cytoolz.map)  # noqa: A001
mapcat = humpy_cytoolz.curry(humpy_cytoolz.mapcat)
nth = humpy_cytoolz.curry(humpy_cytoolz.nth)
partial = humpy_cytoolz.curry(humpy_cytoolz.partial)
partition = humpy_cytoolz.curry(humpy_cytoolz.partition)
partition_all = humpy_cytoolz.curry(humpy_cytoolz.partition_all)
partitionby = humpy_cytoolz.curry(humpy_cytoolz.partitionby)
peekn = humpy_cytoolz.curry(humpy_cytoolz.peekn)
pluck = humpy_cytoolz.curry(humpy_cytoolz.pluck)
random_sample = humpy_cytoolz.curry(humpy_cytoolz.random_sample)
reduce = humpy_cytoolz.curry(humpy_cytoolz.reduce)
reduceby = humpy_cytoolz.curry(humpy_cytoolz.reduceby)
remove = humpy_cytoolz.curry(humpy_cytoolz.remove)
sliding_window = humpy_cytoolz.curry(humpy_cytoolz.sliding_window)
sorted = humpy_cytoolz.curry(humpy_cytoolz.sorted)  # noqa: A001
tail = humpy_cytoolz.curry(humpy_cytoolz.tail)
take = humpy_cytoolz.curry(humpy_cytoolz.take)
take_nth = humpy_cytoolz.curry(humpy_cytoolz.take_nth)
topk = humpy_cytoolz.curry(humpy_cytoolz.topk)
unique = humpy_cytoolz.curry(humpy_cytoolz.unique)
update_in = humpy_cytoolz.curry(humpy_cytoolz.update_in)
valfilter = humpy_cytoolz.curry(humpy_cytoolz.valfilter)
valmap = humpy_cytoolz.curry(humpy_cytoolz.valmap)
del humpy_cytoolz



from .dev_skip_test import dev_skip_test
import humpy_cytoolz
import humpy_cytoolz.curried
import types

@dev_skip_test
def test_toolzcurry_is_class():
    import humpy_toolz
    assert isinstance(humpy_toolz.curry, type) is True
    assert isinstance(humpy_toolz.curry, types.FunctionType) is False

@dev_skip_test
def test_cytoolz_like_toolz():
    import humpy_toolz
    import humpy_toolz.curried
    for key, val in vars(humpy_toolz.curried).items():
        if isinstance(val, humpy_toolz.curry):
            if val.func is humpy_toolz.curry:
                continue
            assert hasattr(humpy_cytoolz.curried, key), 'cytoolz.curried.%s does not exist' % key
            assert isinstance(getattr(humpy_cytoolz.curried, key), humpy_cytoolz.curry), 'cytoolz.curried.%s should be curried' % key

@dev_skip_test
def test_toolz_like_cytoolz():
    import humpy_toolz
    import humpy_toolz.curried
    for key, val in vars(humpy_cytoolz.curried).items():
        if isinstance(val, humpy_cytoolz.curry):
            assert hasattr(humpy_toolz.curried, key), 'cytoolz.curried.%s should not exist' % key
            assert isinstance(getattr(humpy_toolz.curried, key), humpy_toolz.curry), 'cytoolz.curried.%s should not be curried' % key



from humpy_cytoolz import *
from humpy_cytoolz.utils import raises
import humpy_cytoolz
import humpy_cytoolz.curried
import pickle

def test_compose():
    f = compose(str, sum)
    g = pickle.loads(pickle.dumps(f))
    assert f((1, 2)) == g((1, 2))

def test_curry():
    f = curry(map)(str)
    g = pickle.loads(pickle.dumps(f))
    assert list(f((1, 2, 3))) == list(g((1, 2, 3)))

def test_juxt():
    f = juxt(str, int, bool)
    g = pickle.loads(pickle.dumps(f))
    assert f(1) == g(1)
    assert f.funcs == g.funcs

def test_complement():
    f = complement(bool)
    assert f(True) is False
    assert f(False) is True
    g = pickle.loads(pickle.dumps(f))
    assert f(True) == g(True)
    assert f(False) == g(False)

def test_instanceproperty():
    p = humpy_cytoolz.functoolz.InstanceProperty(bool)
    assert p.__get__(None) is None
    assert p.__get__(0) is False
    assert p.__get__(1) is True
    p2 = pickle.loads(pickle.dumps(p))
    assert p2.__get__(None) is None
    assert p2.__get__(0) is False
    assert p2.__get__(1) is True

def f(x, y):
    return (x, y)

def test_flip():
    flip = pickle.loads(pickle.dumps(humpy_cytoolz.functoolz.flip))
    assert flip is humpy_cytoolz.functoolz.flip
    g1 = flip(f)
    g2 = pickle.loads(pickle.dumps(g1))
    assert g1(1, 2) == g2(1, 2) == f(2, 1)
    g1 = flip(f)(1)
    g2 = pickle.loads(pickle.dumps(g1))
    assert g1(2) == g2(2) == f(2, 1)

def test_curried_exceptions():
    merge = pickle.loads(pickle.dumps(humpy_cytoolz.curried.merge))
    assert merge is humpy_cytoolz.curried.merge

@humpy_cytoolz.curry
class GlobalCurried:

    def __init__(self, x, y):
        self.x = x
        self.y = y

    @humpy_cytoolz.curry
    def f1(self, a, b):
        return self.x + self.y + a + b

    def g1(self):
        pass

    def __reduce__(self):
        """Allow us to serialize instances of GlobalCurried"""
        return (GlobalCurried, (self.x, self.y))

    @humpy_cytoolz.curry
    class NestedCurried:

        def __init__(self, x, y):
            self.x = x
            self.y = y

        @humpy_cytoolz.curry
        def f2(self, a, b):
            return self.x + self.y + a + b

        def g2(self):
            pass

        def __reduce__(self):
            """Allow us to serialize instances of NestedCurried"""
            return (GlobalCurried.NestedCurried, (self.x, self.y))

    class Nested:

        def __init__(self, x, y):
            self.x = x
            self.y = y

        @humpy_cytoolz.curry
        def f3(self, a, b):
            return self.x + self.y + a + b

        def g3(self):
            pass

def test_curried_qualname():

    def preserves_identity(obj):
        return pickle.loads(pickle.dumps(obj)) is obj
    assert preserves_identity(GlobalCurried)
    assert preserves_identity(GlobalCurried.func.f1)
    assert preserves_identity(GlobalCurried.func.NestedCurried)
    assert preserves_identity(GlobalCurried.func.NestedCurried.func.f2)
    assert preserves_identity(GlobalCurried.func.Nested.f3)
    global_curried1 = GlobalCurried(1)
    global_curried2 = pickle.loads(pickle.dumps(global_curried1))
    assert global_curried1 is not global_curried2
    assert global_curried1(2).f1(3, 4) == global_curried2(2).f1(3, 4) == 10
    global_curried3 = global_curried1(2)
    global_curried4 = pickle.loads(pickle.dumps(global_curried3))
    assert global_curried3 is not global_curried4
    assert global_curried3.f1(3, 4) == global_curried4.f1(3, 4) == 10
    func1 = global_curried1(2).f1(3)
    func2 = pickle.loads(pickle.dumps(func1))
    assert func1 is not func2
    assert func1(4) == func2(4) == 10
    nested_curried1 = GlobalCurried.func.NestedCurried(1)
    nested_curried2 = pickle.loads(pickle.dumps(nested_curried1))
    assert nested_curried1 is not nested_curried2
    assert nested_curried1(2).f2(3, 4) == nested_curried2(2).f2(3, 4) == 10

def test_curried_bad_qualname():

    @humpy_cytoolz.curry
    class Bad:
        __qualname__ = 'cytoolz.functoolz.not.a.valid.path'
    assert raises(pickle.PicklingError, lambda: pickle.dumps(Bad))



from humpy_cytoolz.functoolz import curry
import operator

IGNORE = {'__abs__', '__index__', '__inv__', '__invert__', '__neg__', '__not__', '__pos__', '_abs', 'abs', 'attrgetter', 'index', 'inv', 'invert', 'is_none', 'is_not_none', 'itemgetter', 'neg', 'not_', 'pos', 'truth'}
locals().update({name: f if name in IGNORE else curry(f) for name, f in vars(operator).items() if callable(f)})
del curry
del operator



from .utils import no_default
from _typeshed import SupportsRichComparison
from collections.abc import Callable, Hashable, Iterable, Iterator, Mapping, Sequence
from typing import Any, Literal, overload, Protocol, TypeGuard

__all__ = (
	"accumulate",
	"concat",
	"concatv",
	"cons",
	"count",
	"diff",
	"drop",
	"first",
	"frequencies",
	"get",
	"groupby",
	"interleave",
	"interpose",
	"isdistinct",
	"isiterable",
	"iterate",
	"join",
	"last",
	"mapcat",
	"merge_sorted",
	"nth",
	"partition",
	"partition_all",
	"peek",
	"peekn",
	"pluck",
	"random_sample",
	"reduceby",
	"remove",
	"second",
	"sliding_window",
	"tail",
	"take",
	"take_nth",
	"topk",
	"unique",
)

### Special types for toolz
type _NoDefaultType = Literal[__no_default__]
type _NoPadType = Literal[__no_pad__]

class _Randomable(Protocol):
	def random(self) -> float: ...

### Toolz itself

def remove[T](
	predicate: Callable[[T], bool], seq: Iterable[T]
) -> Iterable[T]:
	...

@overload
def accumulate[T](
	binop: Callable[[T, T], T],
	seq: Iterable[T],
) -> Iterator[T]: ...
@overload
def accumulate[T](
	binop: Callable[[T, T], T], seq: Iterable[T], initial: T
) -> Iterator[T]: ...
def accumulate[T](
	binop: Callable[[T, T], T],
	seq: Iterable[T],
	initial: T | _NoDefaultType = no_default,
) -> Iterator[T]:
	...

def groupby[KT, T](
	key: Callable[[T], KT], seq: Iterable[T]
) -> dict[KT, list[T]]:
	...

def merge_sorted[CT: SupportsRichComparison](
	*seqs: Iterable[CT], key: Callable[[CT], CT] | None = None
) -> Iterator[CT]:
	...

def interleave[T](
	seqs: Iterable[Iterable[T]],
) -> Iterator[T]:
	...

def unique[T](
	seq: Iterable[T],
	key: Callable[[T], Any] | None = None,
) -> Iterator[T]:
	...

def isiterable(x: Any) -> TypeGuard[Iterable[Any]]:
	...

def isdistinct(
	seq: Iterable[Any] | Sequence[Any],
) -> bool:
	...

def take[T](n: int, seq: Iterable[T]) -> Iterator[T]:
	...

def tail[T](n: int, seq: Iterable[T]) -> Iterator[T]:
	...

def drop[T](n: int, seq: Iterable[T]) -> Iterator[T]:
	...

def take_nth[T](
	n: int, seq: Iterable[T]
) -> Iterator[T]:
	...

def first[T](seq: Iterable[T]) -> T:
	...

def second[T](seq: Iterable[T]) -> T:
	...

def nth[T](n: int, seq: Iterable[T]) -> T:
	...

def last[T](seq: Iterable[T]) -> T:
	...

def rest[T](seq: Iterable[T]) -> Iterable[T]:
	...
	# Warning - this function is not exposed via __all__ and should be considered private.

@overload
def get[T](
	ind: list[Any],
	seq: Sequence[T] | Mapping[Any, T],
	default: T | _NoDefaultType = ...,
) -> tuple[T, ...]: ...
@overload
def get[T](
	ind: Any,
	seq: Sequence[T] | Mapping[Any, T],
	default: T | _NoDefaultType = ...,
) -> T: ...
def get[T](
	ind: Any | list[Any],
	seq: Sequence[T] | Mapping[Any, T],
	default: T | _NoDefaultType = no_default,
) -> T | tuple[T, ...]:
	...

def concat[T](
	seqs: Iterable[Iterable[T]],
) -> Iterator[T]:
	...

def concatv[T](*seqs: Iterable[T]) -> Iterator[T]:
	...

def mapcat[T, R](
	func: Callable[[T], Iterable[R]],
	seqs: Iterable[T],
) -> Iterator[R]:
	...

def cons[T](el: T, seq: Iterable[T]) -> Iterator[T]:
	...

def interpose[T](
	el: T, seq: Iterable[T]
) -> Iterator[T]:
	...

def frequencies[T](seq: Iterable[T]) -> dict[T, int]:
	...

@overload
def reduceby[T, K](
	key: Callable[[T], K],
	binop: Callable[[T, T], T],
	seq: Iterable[T],
) -> dict[K, T]: ...
@overload
def reduceby[T, K](
	key: Callable[[T], K],
	binop: Callable[[T, T], T],
	seq: Iterable[T],
	init: T | Callable[[], T],
) -> dict[K, T]: ...
@overload
def reduceby[T](
	key: Any,  # when not callable, use identity function
	binop: Callable[[T, T], T],
	seq: Iterable[T],
) -> dict[T, T]: ...
@overload
def reduceby[T](
	key: Any,  # when not callable, use identity function
	binop: Callable[[T, T], T],
	seq: Iterable[T],
	init: T | Callable[[], T],
) -> dict[T, T]: ...
def reduceby[T, K](
	key: Callable[[T], K] | Any,
	binop: Callable[[T, T], T],
	seq: Iterable[T],
	init: T | Callable[[], T] | _NoDefaultType = no_default,
) -> dict[K, T]:
	...

def iterate[T](func: Callable[[T], T], x: T) -> Iterator[T]:
	...

def sliding_window[T](
	n: int, seq: Iterable[T]
) -> Iterator[tuple[T, ...]]:
	...

no_pad = "__no_pad__"

@overload
def partition[T, P](
	n: Literal[1], seq: Iterable[T], pad: Any = ...
) -> Iterator[tuple[T]]: ...
@overload
def partition[T](
	n: int, seq: Iterable[T], pad: _NoPadType = ...
) -> Iterator[tuple[T, ...]]: ...
@overload
def partition[T, P](
	n: int, seq: Iterable[T], pad: P
) -> Iterator[tuple[T | P, ...]]: ...
def partition[T, P](
	n: int, seq: Iterable[T], pad: P | _NoPadType = no_pad
) -> Iterator[tuple[T | P, ...]]:
	...

@overload
def partition_all[T](
	n: Literal[1], seq: Iterable[T]
) -> Iterator[tuple[T]]: ...
@overload
def partition_all[T](
	n: int, seq: Iterable[T]
) -> Iterator[tuple[T, ...]]: ...
def partition_all[T](
	n: int, seq: Iterable[T]
) -> Iterator[tuple[T, ...]]:
	...

def count(seq: Iterable[Any]) -> int:
	...

@overload
def pluck[T](
	ind: list[Any],
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	],
	default: T | _NoDefaultType = ...,
) -> Iterator[tuple[T, ...]]: ...
@overload
def pluck[T](
	ind: Any,
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	],
	default: T | _NoDefaultType = ...,
) -> Iterator[T]: ...
def pluck[T](
	ind: Any | list[Any],
	seqs: Iterable[
		Sequence[T] | Mapping[Any, T]
	],
	default: T | _NoDefaultType = no_default,
) -> Iterator[T] | Iterator[tuple[T, ...]]:
	...

@overload
def getter[T](
	index: list[Any],
) -> Callable[
	[Sequence[T] | Mapping[Any, T]],
	tuple[T, ...],
]: ...
@overload
def getter[T](
	index: Any,
) -> Callable[
	[Sequence[T] | Mapping[Any, T]], T
]: ...
def getter[T](
	index: Any | list[Any],
) -> Callable[
	[Sequence[T] | Mapping[Any, T]],
	T | tuple[T, ...],
]:
	# Warning - this function is not exposed via __all__ and should be considered private.
	...

# === CALLABLE + CALLABLE (4 overloads) ===
@overload
def join[T, U](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
) -> Iterator[tuple[T, U]]: ...
@overload
def join[T, U, L](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	left_default: L,
) -> Iterator[tuple[T | L, U]]: ...
@overload
def join[T, U, R](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	*,
	right_default: R,
) -> Iterator[tuple[T, U | R]]: ...
@overload
def join[T, U, L, R](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	left_default: L,
	right_default: R,
) -> Iterator[tuple[T | L, U | R]]: ...

# === HASHABLE + CALLABLE (4 overloads) ===
@overload
def join[T, U](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
) -> Iterator[tuple[T, U]]: ...
@overload
def join[T, U, L](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	left_default: L,
) -> Iterator[tuple[T | L, U]]: ...
@overload
def join[T, U, R](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	*,
	right_default: R,
) -> Iterator[tuple[T, U | R]]: ...
@overload
def join[T, U, L, R](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable],
	rightseq: Iterable[U],
	left_default: L,
	right_default: R,
) -> Iterator[tuple[T | L, U | R]]: ...

# === CALLABLE + HASHABLE (4 overloads) ===
@overload
def join[T, U](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
) -> Iterator[tuple[T, U]]: ...
@overload
def join[T, U, L](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
	left_default: L,
) -> Iterator[tuple[T | L, U]]: ...
@overload
def join[T, U, R](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
	*,
	right_default: R,
) -> Iterator[tuple[T, U | R]]: ...
@overload
def join[T, U, L, R](
	leftkey: Callable[[T], Hashable],
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
	left_default: L,
	right_default: R,
) -> Iterator[tuple[T | L, U | R]]: ...

# === HASHABLE + HASHABLE (4 overloads) ===
@overload
def join[T, U](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
) -> Iterator[tuple[T, U]]: ...
@overload
def join[T, U, L](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
	left_default: L,
) -> Iterator[tuple[T | L, U]]: ...
@overload
def join[T, U, R](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
	*,
	right_default: R,
) -> Iterator[tuple[T, U | R]]: ...
@overload
def join[T, U, L, R](
	leftkey: Hashable,
	leftseq: Iterable[T],
	rightkey: Hashable,
	rightseq: Iterable[U],
	left_default: L,
	right_default: R,
) -> Iterator[tuple[T | L, U | R]]: ...

# Implementation signature
def join[T, U, L, R](
	leftkey: Callable[[T], Hashable] | Hashable,
	leftseq: Iterable[T],
	rightkey: Callable[[U], Hashable] | Hashable,
	rightseq: Iterable[U],
	left_default: L | _NoDefaultType = no_default,
	right_default: R | _NoDefaultType = no_default,
) -> Iterator[tuple[T | L, U | R]]:
	...

@overload
def diff[T](
	*seqs: Iterable[T],
	key: Callable[[T], Any] | None = None,
) -> Iterator[tuple[T, ...]]: ...
@overload
def diff[T](
	*seqs: Iterable[T],
	default: T,
	key: Callable[[T], Any] | None = None,
) -> Iterator[tuple[T, ...]]: ...
def diff[T](
	*seqs: Iterable[T],
	default: T | _NoDefaultType = no_default,
	key: Callable[[T], Any] | None = None,
) -> Iterator[tuple[T, ...]]:
	...

def topk[T](
	k: int,
	seq: Iterable[T],
	key: Callable[[T], SupportsRichComparison] | None = None,
) -> tuple[T, ...]:
	...

def peek[T](
	seq: Iterable[T],
) -> tuple[T, Iterator[T]]:
	...

def peekn[T](
	n: int, seq: Iterable[T]
) -> tuple[tuple[T, ...], Iterator[T]]:
	...

def random_sample[T](
	prob: float,
	seq: Iterable[T],
	random_state: int | _Randomable | None = None,
) -> Iterator[T]:
	...



"""Access comprehensive test suites for the hunterMakesPy package.

(AI generated docstring)

You can use this package to run test suites that validate all major functions and features
in hunterMakesPy. The tests cover defensive programming utilities, parameter validation,
file system operations, and data structure manipulation. Each test module corresponds to a
public module in the parent package and can be executed independently using pytest [1].

The test suites use the conftest.py infrastructure for shared fixtures and test utilities.
Tests are organized by module with comprehensive coverage of edge cases, error conditions,
and typical usage patterns. You can import individual test functions or run entire test
modules to validate behavior in your environment.

Test Modules
------------
test_coping
	Tests for defensive programming utilities and package configuration.
test_dataStructures
	Tests for data structure manipulation and encoding functions.
test_filesystemToolkit
	Tests for file system operations and dynamic imports.
test_parseParameters
	Tests for parameter validation and concurrency limit computation.
test_theTypes
	Tests for type protocols and type aliases.

References
----------
[1] pytest - Context7
	https://docs.pytest.org/

"""



"""``humpy_tlz`` mirrors the ``humpy_toolz`` API and uses ``humpy_cytoolz`` if possible.

The ``humpy_tlz`` package is installed when ``humpy_toolz`` is installed.  It provides
a convenient way to use functions from ``humpy_cytoolz``--a faster Cython
implementation of ``humpy_toolz``--if it is installed, otherwise it uses
functions from ``humpy_toolz``.
"""
from . import _build_tlz




import doctest
import humpy_cytoolz
import humpy_cytoolz.dicttoolz
import humpy_cytoolz.functoolz
import humpy_cytoolz.itertoolz
import humpy_cytoolz.recipes

def module_doctest(m, *args, **kwargs):
    return doctest.testmod(m, *args, **kwargs).failed == 0

def test_doctest():
    assert module_doctest(humpy_cytoolz)
    assert module_doctest(humpy_cytoolz.dicttoolz)
    assert module_doctest(humpy_cytoolz.functoolz)
    assert module_doctest(humpy_cytoolz.itertoolz)
    assert module_doctest(humpy_cytoolz.recipes)



# cython: embedsignature=True
# cython: freethreading_compatible=True
# cython: language_level=3
import os.path
import humpy_cytoolz


__all__ = ['raises', 'no_default', 'include_dirs', 'consume']


try:
    # Attempt to get the no_default sentinel object from humpy_toolz
    from humpy_toolz.utils import no_default
except ImportError:
    no_default = '__no__default__'


def raises(err, lamda):
    try:
        lamda()
        return False
    except err:
        return True


def include_dirs():
    """ Return a list of directories containing the *.pxd files for ``humpy_cytoolz``

    Use this to include ``humpy_cytoolz`` in your own Cython project, which allows
    fast C bindinds to be imported such as ``from humpy_cytoolz cimport get``.

    Below is a minimal "setup.py" file using ``include_dirs``:

        from setuptools import setup
        from setuptools.extension import Extension
        from Cython.Build import cythonize

        import humpy_cytoolz.utils

        ext_modules=[
            Extension("mymodule",
                      ["mymodule.pyx"],
                      include_dirs=humpy_cytoolz.utils.include_dirs()
                     )
        ]
        ext_modules = cythonize(ext_modules)

        setup(
          name="mymodule",
          ext_modules=ext_modules,
        )
    """
    return os.path.split(humpy_cytoolz.__path__[0])


cpdef object consume(object seq):
    """
    Efficiently consume an iterable """
    for _ in seq:
        pass



"""Provide immutable, functional-style operations on `Mapping`[1] objects.

(AI generated docstring)

You can use this module to transform, filter, retrieve from, and merge `Mapping`[1] objects
without modifying the source mapping. Each function returns a new mapping created by a
`factory` argument that defaults to `dict`. Functions that navigate nested structures accept
a `Sequence`[1] of keys to specify the path.

Contents
--------
Classes
	SupportsGetItem
		Structural type constraint for objects that support item retrieval via `__getitem__`.

Functions
	assoc
		Create a new `Mapping` with `key` associated with `value`.
	assoc_in
		Create a new `MutableMapping` from `d` with `value` at the path specified by `keys`.
	dissoc
		Create a new `MutableMapping` from `d` with the specified `keys` removed.
	get_in
		Retrieve a value from a potentially nested collection using a sequence of keys.
	itemfilter
		Retain only items from `d` whose key-value pairs satisfy `predicate` and return a new `Mapping`.
	itemmap
		Apply `func` to all items of `d` and return a new `Mapping` with the transformed items.
	keyfilter
		Retain only items from `d` whose keys satisfy `predicate` and return a new `Mapping`.
	keymap
		Apply `func` to all keys of `d` and return a new `Mapping` with the transformed keys.
	merge
		Merge a collection of `Mapping` objects and return a new `Mapping`.
	merge_with
		Merge `Mapping` objects and apply a `Callable` to combined values.
	update_in
		Apply a `Callable` to a value at a nested path in a `Mapping`.
	valfilter
		Retain only items from `d` whose values satisfy `predicate` and return a new `Mapping`.
	valmap
		Apply `func` to all values of `d` and return a new `Mapping` with the transformed values.

References
----------
[1] Python `collections.abc` module
	https://docs.python.org/3/library/collections.abc.html
"""
from collections import defaultdict, deque
from collections.abc import Callable, Hashable, Mapping, MutableMapping, Sequence
from functools import reduce
from typing import Any, Literal, overload, Protocol, TypeGuard
from typing_extensions import TypeIs
import operator

__all__ = ('assoc', 'assoc_in', 'dissoc', 'get_in', 'itemfilter', 'itemmap', 'keyfilter', 'keymap', 'merge', 'merge_with', 'update_in', 'valfilter', 'valmap')

class SupportsGetItem[K: Hashable, V](Protocol):
	def __getitem__(self, key: K, /) -> V: ...

@overload
def assoc[K: Hashable, V](d: Mapping[K, V], key: K, value: V, factory: Callable[[], dict[K, V]] = dict) -> dict[K, V]: ...
@overload
def assoc[K: Hashable, V](d: Mapping[K, V], key: K, value: V, factory: Callable[[], MutableMapping[K, V]]) -> MutableMapping[K, V]: ...
def assoc[K: Hashable, V](d: Mapping[K, V], key: K, value: V, factory: Callable[[], MutableMapping[K, V]] = dict) -> MutableMapping[K, V]:
	"""Create a new `Mapping`[1] with `key` associated with `value`.

	You can use `assoc` (***assoc***iate) to copy `d` (***d***ictionary) to a new `Mapping` created by
	`factory` and assign `value` to `key`. `assoc` does not change `d`.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	key : K
		`key` that `assoc` inserts or replaces.
	value : V
		`value` that `assoc` assigns to `key`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingUpdated : MutableMapping[K, V]
		New `Mapping` with `key` associated to `value`.

	Examples
	--------
	>>> assoc({}, 'a', 1)
	{'a': 1}
	>>> assoc({'a': 1}, 'a', 3)
	{'a': 3}
	>>> assoc({'a': 1}, 'b', 3)
	{'a': 1, 'b': 3}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	d2: MutableMapping[K, V] = factory()
	d2.update(d)
	d2[key] = value
	return d2

# Overloads for nested dictionaries with tuple keys (2-level nesting)
@overload
def assoc_in[K1, K2, V1, V2](d: Mapping[K1, Mapping[K2, V2] | V1], keys: tuple[K1, K2], value: V2) -> dict[K1, dict[K2, V2] | V1 | V2]: ...
@overload
def assoc_in[K1, K2, V1, V2](d: Mapping[K1, Mapping[K2, V2] | V1], keys: tuple[K1, K2], value: V2, *, factory: Callable[ [], MutableMapping[K1, Any] ]) -> MutableMapping[K1, Any]: ...

# Overloads for nested dictionaries with tuple keys (3-level nesting)
@overload
def assoc_in[K1, K2, K3, V1, V2, V3](d: Mapping[ K1, Mapping[K2, Mapping[K3, V3] | V2] | V1 ], keys: tuple[K1, K2, K3], value: V3) -> dict[K1, dict[K2, dict[K3, V3] | V2 | V3] | V1 | V3]: ...
@overload
def assoc_in[K1, K2, K3, V1, V2, V3](d: Mapping[ K1, Mapping[K2, Mapping[K3, V3] | V2] | V1 ], keys: tuple[K1, K2, K3], value: V3, *, factory: Callable[ [], MutableMapping[K1, Any] ]) -> MutableMapping[K1, Any]: ...

# General overloads for backwards compatibility
@overload
def assoc_in[K, V](d: Mapping[K, V], keys: Sequence[K], value: V) -> dict[K, V]: ...
@overload
def assoc_in[K, V](d: Mapping[K, V], keys: Sequence[K], value: V, *, factory: Callable[[], MutableMapping[K, V]]) -> MutableMapping[K, V]: ...
def assoc_in[K, V](d: Mapping[K, V], keys: Sequence[K], value: V, *, factory: Callable[[], MutableMapping[K, V]] = dict) -> MutableMapping[K, V]: # pyright: ignore[reportInconsistentOverload]
	"""Create a new `MutableMapping` from `d` with `value` at the path specified by `keys`.

	(AI generated docstring)

	You can use `assoc_in` to produce a copy of `d` with `value` placed at the nested
	location given by `keys`. `assoc_in` creates intermediate `MutableMapping` instances
	as needed when a key in `keys` is absent from `d`. `assoc_in` does not mutate `d` or
	any nested `Mapping` within `d`.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	keys : Iterable[K]
		Non-empty sequence of keys specifying the nested path to the target location in `d`.
	value : V
		The value to place at the location specified by `keys`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates each new `MutableMapping`[1] in the result.

	Returns
	-------
	mappingUpdated : MutableMapping[K, V]
		New `MutableMapping` based on `d` with `value` at the path specified by `keys`.

	See Also
	--------
	update_in : Apply a `Callable` to a value at a nested path in a `Mapping`.
	get_in : Retrieve a value at a nested path in a `Mapping`.
	assoc : Create a new `Mapping` from `d` with one key associated to a value.

	Examples
	--------
	>>> assoc_in({'a': 1}, ['a'], 2)
	{'a': 2}
	>>> assoc_in({'a': {'b': 1}}, ['a', 'b'], 2)
	{'a': {'b': 2}}
	>>> assoc_in({}, ['a', 'b'], 1)
	{'a': {'b': 1}}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	return update_in(d, keys, lambda _x: value, value, factory)  # pyright: ignore[reportUnknownVariableType, reportCallIssue]

@overload
def dissoc[K: Hashable, V](d: Mapping[K, V], *keys: K, factory: Callable[[], dict[K, V]]) -> dict[K, V]: ...
@overload
def dissoc[K: Hashable, V](d: Mapping[K, V], *keys: K, factory: Callable[[], MutableMapping[K, V]]) -> MutableMapping[K, V]: ...
def dissoc[K: Hashable, V](d: Mapping[K, V], *keys: K, factory: Callable[[], MutableMapping[K, V]] = dict) -> MutableMapping[K, V]:
	"""Create a new `MutableMapping`[1] from `d` with the specified `keys` removed.

	(AI generated docstring)

	You can use `dissoc` (***dissoc***iate) to copy `d` (***d***ictionary) to a new `MutableMapping`
	created by `factory`, then remove each key in `keys` from the result. `dissoc` does not change
	`d`. Keys in `keys` that are absent from `d` are silently ignored.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	*keys : K
		Keys to remove from `d` in the returned `MutableMapping`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingReduced : MutableMapping[K, V]
		New `MutableMapping` containing all items from `d` except those whose keys are in `keys`.

	Algorithm Details
	-----------------
	`dissoc` selects between two strategies based on the ratio of removed keys to total keys.

	When fewer than 60% of keys are removed, `dissoc` copies all items from `d` and then
	deletes the specified keys one by one.

	When 60% or more of keys are removed, `dissoc` computes the set of remaining keys and
	copies only those items to the result, avoiding unnecessary copy-and-delete operations.

	See Also
	--------
	assoc : Create a new `MutableMapping` from `d` with one key associated to a value.

	Examples
	--------
	>>> dissoc({'x': 1, 'y': 2}, 'y')
	{'x': 1}
	>>> dissoc({'x': 1, 'y': 2}, 'y', 'x')
	{}
	>>> dissoc({'x': 1}, 'y') # Ignores missing keys
	{'x': 1}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	d2: MutableMapping[K, V] = factory()
	if len(keys) < len(d) * 0.6:
		d2.update(d)
		for key in keys:
			if key in d2:
				del d2[key]
	else:
		remaining: set[K] = set(d)
		remaining.difference_update(keys)
		for k in remaining:
			d2[k] = d[k]
	return d2

@overload
def get_in[K: Hashable, V](keys: Sequence[K], coll: SupportsGetItem[K, V], default: None = None, no_default: Literal[True] = True) -> V: ...  # noqa: FBT002
@overload
def get_in[K: Hashable, V](keys: Sequence[K], coll: SupportsGetItem[K, V], default: None = None, no_default: bool = False) -> V | None: ...  # noqa: FBT001, FBT002
@overload
def get_in[K: Hashable, V](keys: Sequence[K], coll: SupportsGetItem[K, V], default: V, no_default: bool = False) -> V: ...  # noqa: FBT001, FBT002
def get_in[K: Hashable, V](keys: Sequence[K], coll: SupportsGetItem[K, V], default: V | None = None, no_default: bool = False) -> V | None:  # noqa: FBT001, FBT002
	"""Retrieve a value from a potentially nested `coll` (***coll***ection) using a `Sequence` of `keys`.

	(AI generated docstring)

	You can use `get_in` to navigate into a nested `coll` (***coll***ection) by following a
	sequence of `keys`. `get_in` applies each key in `keys` sequentially using
	`operator.getitem`[1]. If the path does not exist, `get_in` returns `default`. If
	`no_default` is `True`, `get_in` re-raises the original exception instead of returning
	`default`.

	Parameters
	----------
	keys : Sequence[K]
		Sequence of keys that describes the path to traverse in `coll`.
	coll : SupportsGetItem[K, V]
		Collection to traverse. `get_in` applies each key in `keys` to the current `coll`
		using `operator.getitem`[1], so `coll` can be any nested structure such as a `dict`
		or `list`.
	default : V | None = None
		Value to return when the path in `keys` does not exist in `coll`.
	no_default : bool = False
		When `True`, re-raise the original `KeyError`, `IndexError`, or `TypeError`
		instead of returning `default`.

	Returns
	-------
	value : V | None
		The value at the nested path in `coll`, or `default` if the path does not exist.

	Raises
	------
	KeyError
		When `no_default` is `True` and a key is missing from a mapping.
	IndexError
		When `no_default` is `True` and an index is out of range.
	TypeError
		When `no_default` is `True` and a key type is incompatible with `coll`.

	See Also
	--------
	itertoolz.get : Retrieve a value or values from a collection.
	operator.getitem : Return the value at a given key in a collection.
	assoc_in : Create a new `Mapping` from `d` with a value at a nested path.
	update_in : Apply a `Callable` to a value at a nested path in a `Mapping`.

	Examples
	--------
	>>> transaction = {'name': 'Alice',
	...                'purchase': {'items': ['Apple', 'Orange'],
	...                             'costs': [0.50, 1.25]},
	...                'credit card': '5555-1234-1234-1234'}
	>>> get_in(['purchase', 'items', 0], transaction)
	'Apple'
	>>> get_in(['name'], transaction)
	'Alice'
	>>> get_in(['purchase', 'total'], transaction)
	>>> get_in(['purchase', 'items', 'apple'], transaction)
	>>> get_in(['purchase', 'items', 10], transaction)
	>>> get_in(['purchase', 'total'], transaction, 0)
	0
	>>> get_in(['y'], {}, no_default=True)
	Traceback (most recent call last):
		...
	KeyError: 'y'

	References
	----------
	[1] Python `operator` module
		https://docs.python.org/3/library/operator.html#operator.getitem
	"""
	try:
		return reduce(operator.getitem, keys, coll) # pyright: ignore[reportReturnType, reportArgumentType]  # ty:ignore[invalid-return-type]
	except (KeyError, IndexError, TypeError):
		if no_default:
			raise
		return default

@overload
def itemfilter[K0: Hashable, V0, K1: Hashable, V1](predicate: Callable[[tuple[K0, V0]], bool], d: Mapping[K0, V0], factory: Callable[[], dict[K1, V1]]) -> dict[K1, V1]: ...
@overload
def itemfilter[K0: Hashable, V0, K1: Hashable, V1](predicate: Callable[[tuple[K0, V0]], TypeGuard[tuple[K1, V1]]], d: Mapping[K0, V0], factory: Callable[[], dict[K1, V1]]) -> dict[K1, V1]: ...
@overload
def itemfilter[K0: Hashable, V0, K1: Hashable, V1](predicate: Callable[[tuple[K0, V0]], TypeIs[tuple[K1, V1]]], d: Mapping[K0, V0], factory: Callable[[], dict[K1, V1]]) -> dict[K1, V1]: ...
@overload
def itemfilter[K0: Hashable, V0, K1: Hashable, V1](predicate: Callable[[tuple[K0, V0]], bool], d: Mapping[K0, V0], factory: Callable[[], MutableMapping[K1, V1]]) -> MutableMapping[K1, V1]: ...
@overload
def itemfilter[K0: Hashable, V0, K1: Hashable, V1](predicate: Callable[[tuple[K0, V0]], TypeGuard[tuple[K1, V1]]], d: Mapping[K0, V0], factory: Callable[[], MutableMapping[K1, V1]]) -> MutableMapping[K1, V1]: ...
@overload
def itemfilter[K0: Hashable, V0, K1: Hashable, V1](predicate: Callable[[tuple[K0, V0]], TypeIs[tuple[K1, V1]]], d: Mapping[K0, V0], factory: Callable[[], MutableMapping[K1, V1]]) -> MutableMapping[K1, V1]: ...
def itemfilter[K0: Hashable, V0, K1: Hashable, V1](predicate: Callable[[tuple[K0, V0]], bool] | Callable[[tuple[K0, V0]], TypeGuard[tuple[K1, V1]]] | Callable[[tuple[K0, V0]], TypeIs[tuple[K1, V1]]], d: Mapping[K0, V0], factory: Callable[[], MutableMapping[K1, V1]] = dict) -> MutableMapping[K1, V1]:
	"""Retain only items from `d` whose key-value pairs satisfy `predicate` and return a new `Mapping`.

	(AI generated docstring)

	You can use `itemfilter` to select items from `d` (***d***ictionary) by both key and value
	simultaneously. `itemfilter` calls `predicate` with each item yielded by `d.items()`. Each
	item is passed to `predicate` as a `tuple[K, V]`. `itemfilter` inserts each item for which
	`predicate` returns `True` into a new `MutableMapping`[1] created by `factory`. `itemfilter`
	does not change `d`.

	Parameters
	----------
	predicate : Callable[[tuple[K, V]], bool]
		`Callable` applied to each item of `d`. Each item is passed to `predicate` as a
		`tuple[K, V]`, and `predicate` must return `True` for the item to be retained.
	d : Mapping[K, V]
		Source `Mapping`[1]. `itemfilter` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingFiltered : MutableMapping[K, V]
		New `MutableMapping` created by `factory` containing only items from `d` for which
		`predicate` returns `True`.

	See Also
	--------
	keyfilter : Retain only items from `d` whose keys satisfy `predicate` and return a new `Mapping`.
	valfilter : Retain only items from `d` whose values satisfy `predicate` and return a new `Mapping`.
	itemmap : Apply a `Callable` to all items of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> def isvalid(item):
	...     k, v = item
	...     return k % 2 == 0 and v < 4

	>>> d = {1: 2, 2: 3, 3: 4, 4: 5}
	>>> itemfilter(isvalid, d)
	{2: 3}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	rv: MutableMapping[K1, V1] = factory()
	for item in d.items():
		if predicate(item):
			k, v = item
			rv[k] = v  # pyright: ignore[reportArgumentType] # ty:ignore[invalid-assignment]
	return rv

@overload
def itemmap[K0: Hashable, K1: Hashable, V0, V1](func: Callable[[tuple[K0, V0]], tuple[K1, V1]], d: Mapping[K0, V0], factory: Callable[..., dict[K1, V1]] = dict) -> dict[K1, V1]: ...
@overload
def itemmap[K0: Hashable, K1: Hashable, V0, V1](func: Callable[[tuple[K0, V0]], tuple[K1, V1]], d: Mapping[K0, V0], factory: Callable[..., MutableMapping[K1, V1]]) -> MutableMapping[K1, V1]: ...
def itemmap[K0: Hashable, K1: Hashable, V0, V1](func: Callable[[tuple[K0, V0]], tuple[K1, V1]], d: Mapping[K0, V0], factory: Callable[..., MutableMapping[K1, V1]] = dict) -> MutableMapping[K1, V1]:
	"""Apply `func` to all items of `d` and return a new `Mapping` with the transformed items.

	(AI generated docstring)

	You can use `itemmap` to transform all keys and values in `d` simultaneously without
	changing `d`. `itemmap` applies `func` to each item yielded by `d.items()`. Each item
	is passed to `func` as a `tuple[K, V]`, and `func` must return a `tuple[L, W]`. `itemmap`
	inserts each returned `tuple` as a new key-value pair in a `MutableMapping`[1] created
	by `factory`.

	Parameters
	----------
	func : Callable[[tuple[K, V]], tuple[L, W]]
		`Callable` applied to each item of `d`. Each item of `d` is passed to `func` as a
		`tuple[K, V]`, and `func` must return a `tuple[L, W]` containing the new key and value.
	d : Mapping[K, V]
		Source `Mapping`[1]. `itemmap` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[L, W]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingTransformed : MutableMapping[L, W]
		New `MutableMapping` created by `factory` populated with each `tuple[L, W]` returned
		by `func`.

	See Also
	--------
	keymap : Apply a `Callable` to all keys of a `Mapping` and return a new `Mapping`.
	valmap : Apply a `Callable` to all values of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> accountids = {'Alice': 10, 'Bob': 20}
	>>> itemmap(reversed, accountids)  # doctest: +SKIP
	{10: 'Alice', 20: 'Bob'}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	return factory(map(func, d.items()))

@overload
def keyfilter[K0: Hashable, K1: Hashable, V](predicate: Callable[[K0], bool], d: Mapping[K0, V], factory: Callable[[], dict[K1, V]] = dict) -> dict[K1, V]: ...
@overload
def keyfilter[K0: Hashable, K1: Hashable, V](predicate: Callable[[K0], TypeGuard[K1]], d: Mapping[K0, V], factory: Callable[[], dict[K1, V]] = dict) -> dict[K1, V]: ...
@overload
def keyfilter[K0: Hashable, K1: Hashable, V](predicate: Callable[[K0], TypeIs[K1]], d: Mapping[K0, V], factory: Callable[[], dict[K1, V]] = dict) -> dict[K1, V]: ...
@overload
def keyfilter[K0: Hashable, K1: Hashable, V](predicate: Callable[[K0], bool], d: Mapping[K0, V], factory: Callable[[], MutableMapping[K1, V]]) -> MutableMapping[K1, V]: ...
@overload
def keyfilter[K0: Hashable, K1: Hashable, V](predicate: Callable[[K0], TypeGuard[K1]], d: Mapping[K0, V], factory: Callable[[], MutableMapping[K1, V]]) -> MutableMapping[K1, V]: ...
@overload
def keyfilter[K0: Hashable, K1: Hashable, V](predicate: Callable[[K0], TypeIs[K1]], d: Mapping[K0, V], factory: Callable[[], MutableMapping[K1, V]]) -> MutableMapping[K1, V]: ...
def keyfilter[K0: Hashable, K1: Hashable, V](predicate: Callable[[K0], bool] | Callable[[K0], TypeGuard[K1]] | Callable[[K0], TypeIs[K1]], d: Mapping[K0, V], factory: Callable[[], MutableMapping[K1, V]] = dict) -> MutableMapping[K1, V]:
	"""Retain only items from `d` whose keys satisfy `predicate` and return a new `Mapping`.

	(AI generated docstring)

	You can use `keyfilter` to select items from `d` (***d***ictionary) by their keys. `keyfilter`
	calls `predicate` with each key yielded by `d.keys()`. `keyfilter` inserts each item whose
	key causes `predicate` to return `True` into a new `MutableMapping`[1] created by `factory`.
	`keyfilter` does not change `d`.

	Parameters
	----------
	predicate : Callable[[K], bool]
		`Callable` applied to each key of `d`. `keyfilter` keeps each item for which `predicate`
		returns `True`.
	d : Mapping[K, V]
		Source `Mapping`[1]. `keyfilter` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingFiltered : MutableMapping[K, V]
		New `MutableMapping` created by `factory` containing only items from `d` whose keys
		cause `predicate` to return `True`.

	See Also
	--------
	valfilter : Retain only items from `d` whose values satisfy `predicate` and return a new `Mapping`.
	itemfilter : Retain only items from `d` whose key-value pairs satisfy `predicate` and return a new `Mapping`.
	keymap : Apply a `Callable` to all keys of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> iseven = lambda x: x % 2 == 0
	>>> d = {1: 2, 2: 3, 3: 4, 4: 5}
	>>> keyfilter(iseven, d)
	{2: 3, 4: 5}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	rv: MutableMapping[K1, V] = factory()
	for k, v in d.items():
		if predicate(k):
			rv[k] = v  # pyright: ignore[reportArgumentType] # ty:ignore[invalid-assignment]
	return rv

@overload
def keymap[K0: Hashable, K1: Hashable, V](func: Callable[[K0], K1], d: Mapping[K0, V], factory: Callable[[], dict[K1, V]] = dict) -> dict[K1, V]: ...
@overload
def keymap[K0: Hashable, K1: Hashable, V](func: Callable[[K0], K1], d: Mapping[K0, V], factory: Callable[[], MutableMapping[K1, V]]) -> MutableMapping[K1, V]: ...
def keymap[K0: Hashable, K1: Hashable, V](func: Callable[[K0], K1], d: Mapping[K0, V], factory: Callable[[], MutableMapping[K1, V]] = dict) -> MutableMapping[K1, V]:
	"""Apply `func` to all keys of `d` and return a new `Mapping` with the transformed keys.

	(AI generated docstring)

	You can use `keymap` to transform all keys in `d` without changing `d`. `keymap` applies
	`func` to each key yielded by `d.keys()` and builds a new `MutableMapping`[1] created by
	`factory`, associating each transformed key with the corresponding original value of `d`.

	Parameters
	----------
	func : Callable[[K], L]
		`Callable` applied to each key of `d`. Each key of `d` is passed to `func`
		individually, and `func` returns the corresponding transformed key.
	d : Mapping[K, V]
		Source `Mapping`[1]. `keymap` reads all keys from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[L, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingTransformed : MutableMapping[L, V]
		New `MutableMapping` created by `factory` in which each key from `d` is replaced by
		the result of applying `func` to that key, associated with the original value of `d`.

	See Also
	--------
	valmap : Apply a `Callable` to all values of a `Mapping` and return a new `Mapping`.
	itemmap : Apply a `Callable` to all items of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> bills = {'Alice': [20, 15, 30], 'Bob': [10, 35]}
	>>> keymap(str.lower, bills)  # doctest: +SKIP
	{'alice': [20, 15, 30], 'bob': [10, 35]}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	rv: MutableMapping[K1, V] = factory()
	rv.update(zip(map(func, d.keys()), d.values(), strict=True))
	return rv

@overload
def merge[K: Hashable, V](*dicts: Mapping[K, V], factory: Callable[[], dict[K, V]] = dict) -> dict[K, V]: ...
@overload
def merge[K: Hashable, V](*dicts: Mapping[K, V], factory: Callable[[], MutableMapping[K, V]]) -> MutableMapping[K, V]: ...
def merge[K: Hashable, V](*dicts: Mapping[K, V], factory: Callable[[], MutableMapping[K, V]] = dict) -> MutableMapping[K, V]:
	"""Merge a collection of dictionaries and return a new `Mapping`.

	(AI generated docstring)

	You can use `merge` to combine two or more `Mapping`[1] objects into a single new `Mapping`.
	`merge` calls `factory` to create the result, then updates it in order with each `Mapping` in
	`dicts`. When the same key appears in more than one element, the value from the later element
	takes precedence. You can also pass a single `Iterable[Mapping[K, V]]` as the sole positional
	argument instead of multiple `Mapping` arguments.

	Parameters
	----------
	*dicts : Mapping[K, V]
		`Mapping` objects to merge. Alternatively, pass a single `Iterable[Mapping[K, V]]` as the
		sole positional argument.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingMerged : MutableMapping[K, V]
		New `MutableMapping` created by `factory` containing all key-value pairs from `dicts`. For
		duplicate keys, the value from the last `Mapping` in `dicts` that contains the key takes
		precedence.

	See Also
	--------
	merge_with : Merge dictionaries and apply a `Callable` to combined values.

	Examples
	--------
	>>> merge({1: 'one'}, {2: 'two'})
	{1: 'one', 2: 'two'}

	Later dictionaries have precedence

	>>> merge({1: 2, 3: 4}, {3: 3, 4: 4})
	{1: 2, 3: 3, 4: 4}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	if len(dicts) == 1 and not isinstance(dicts[0], Mapping):
		dicts = dicts[0]
	rv: MutableMapping[K, V] = factory()
	for d in dicts:
		rv.update(d)
	return rv

@overload
def merge_with[K: Hashable, V](func: Callable[[Sequence[V]], V], *dicts: Mapping[K, V], factory: Callable[[], dict[K, V]] = dict) -> dict[K, V]: ...
@overload
def merge_with[K: Hashable, V](func: Callable[[Sequence[V]], V], *dicts: Mapping[K, V], factory: Callable[[], MutableMapping[K, V]]) -> MutableMapping[K, V]: ...
def merge_with[K: Hashable, V](func: Callable[[Sequence[V]], V], *dicts: Mapping[K, V], factory: Callable[[], MutableMapping[K, V]] = dict) -> MutableMapping[K, V]:
	"""Merge dictionaries and apply a `Callable` to combined values.

	(AI generated docstring)

	You can use `merge_with` to combine two or more `Mapping`[1] objects and resolve key conflicts
	by applying `func` to a `list` of all values associated with each key. For each key that
	appears in one or more elements of `dicts`, `merge_with` collects all associated values into a
	`list` in the order they appear across `dicts`, then calls `func` with that `list` to produce
	the value in the result. You can also pass a single `Iterable[Mapping[K, V]]` as the sole
	positional argument after `func`.

	Parameters
	----------
	func : Callable[[list[V]], V]
		`Callable` applied to the `list` of values associated with each key across all `Mapping`
		objects in `dicts`. `func` receives a non-empty `list` and must return the merged value for
		that key.
	*dicts : Mapping[K, V]
		`Mapping` objects to merge. Alternatively, pass a single `Iterable[Mapping[K, V]]` as the
		sole positional argument after `func`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingMerged : MutableMapping[K, V]
		New `MutableMapping` created by `factory` where each key maps to the result of calling
		`func` with all values associated with that key across `dicts`.

	See Also
	--------
	merge : Merge a collection of `Mapping` objects into a new `Mapping`.

	Examples
	--------
	>>> merge_with(sum, {1: 1, 2: 2}, {1: 10, 2: 20})
	{1: 11, 2: 22}

	>>> merge_with(first, {1: 1, 2: 2}, {2: 20, 3: 30})  # doctest: +SKIP
	{1: 1, 2: 2, 3: 30}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	if len(dicts) == 1 and not isinstance(dicts[0], Mapping):
		dicts = dicts[0]
	groupedValues: defaultdict[K, list[V]] = defaultdict(list)
	for d in dicts:
		for k, v in d.items():
			groupedValues[k].append(v)
	rv: MutableMapping[K, V] = factory()
	for k, valueList in groupedValues.items():
		rv[k] = func(valueList)
	return rv

# TODO Given `d: dict[str, int | str]`. It does not follow that `func: Callable[[int | str], int | str]`
@overload
def update_in[K, V](d: Mapping[K, V], keys: Sequence[K], func: Callable[[V | None], V], default: None = None, *, factory: Callable[..., MutableMapping[K, V]] = dict) -> dict[K, V]: ...
@overload
def update_in[K, V](d: Mapping[K, V], keys: Sequence[K], func: Callable[[V], V], default: V, factory: Callable[..., MutableMapping[K, V]] = dict) -> dict[K, V]: ...
@overload
def update_in[K, V](d: Mapping[K, V], keys: Sequence[K], func: Callable[[V | None], V], default: None = None, *, factory: Callable[..., MutableMapping[K, V]]) -> MutableMapping[K, V]: ...
@overload
def update_in[K, V](d: Mapping[K, V], keys: Sequence[K], func: Callable[[V], V], default: V, factory: Callable[..., MutableMapping[K, V]]) -> MutableMapping[K, V]: ...
def update_in[K, V](d: Mapping[K, V], keys: Sequence[K], func: Callable[[V], V], default: V | None = None, factory: Callable[..., MutableMapping[K, V]] = dict) -> MutableMapping[K, V]:
	"""Apply a `Callable` to a value at a nested path in a `Mapping`.

	(AI generated docstring)

	You can use `update_in` to produce a copy of `d` with the value at the nested path
	specified by `keys` replaced by the result of calling `func` on the current value.
	`update_in` creates intermediate `MutableMapping` instances as needed when a key in
	`keys` is absent from `d`. If the innermost key is absent, `func` receives `default`
	instead of an existing value. `update_in` does not mutate `d` or any nested `Mapping`
	within `d`.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	keys : Sequence[K]
		Non-empty sequence of keys specifying the nested path to the value to update in `d`.
	func : Callable[[V], V]
		`Callable` applied to the current value at the path in `keys`. If the innermost
		key is absent from `d`, `func` receives `default`.
	default : V | None = None
		Value passed to `func` when the innermost key is absent from `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates each new `MutableMapping`[1] in the result.

	Returns
	-------
	mappingUpdated : MutableMapping[K, V]
		New `MutableMapping` based on `d` with the value at the path specified by `keys`
		replaced by the result of `func`.

	See Also
	--------
	assoc_in : Create a new `Mapping` from `d` with a value at a nested path.
	get_in : Retrieve a value at a nested path in a `Mapping`.
	assoc : Create a new `Mapping` from `d` with one key associated to a value.

	Examples
	--------
	>>> inc = lambda x: x + 1
	>>> update_in({'a': 0}, ['a'], inc)
	{'a': 1}

	>>> update_in({}, ['z'], inc, 0)
	{'z': 1}

	>>> update_in({}, [1, 2, 3], str, default="bar")
	{1: {2: {3: 'bar'}}}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	rv = dATk = factory(d)
	dequeKeys: deque[K] = deque(keys)
	keyFinal: K = dequeKeys.pop()
	sherpa = factory(d)
	while dequeKeys:
		k: K = dequeKeys.popleft()
		sherpa = sherpa.get(k, factory()) # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportUnknownVariableType]  # ty:ignore[unresolved-attribute]
		dATk[k] = dATk = factory(sherpa)  # pyright: ignore[reportArgumentType] # ty:ignore[invalid-assignment]
	dATk[keyFinal] = func(sherpa.get(keyFinal, default))  # pyright: ignore[reportArgumentType, reportUnknownMemberType, reportAttributeAccessIssue] # ty:ignore[unresolved-attribute, invalid-argument-type]
	return rv

@overload
def valfilter[K: Hashable, V0, V1](predicate: Callable[[V0], bool], d: Mapping[K, V0], factory: Callable[[], dict[K, V1]]) -> dict[K, V1]: ...
@overload
def valfilter[K: Hashable, V0, V1](predicate: Callable[[V0], TypeGuard[V1]], d: Mapping[K, V0], factory: Callable[[], dict[K, V1]]) -> dict[K, V1]: ...
@overload
def valfilter[K: Hashable, V0, V1](predicate: Callable[[V0], TypeIs[V1]], d: Mapping[K, V0], factory: Callable[[], dict[K, V1]]) -> dict[K, V1]: ...
@overload
def valfilter[K: Hashable, V0, V1](predicate: Callable[[V0], TypeIs[V1]], d: Mapping[K, V0], factory: Callable[[], MutableMapping[K, V1]]) -> MutableMapping[K, V1]: ...
@overload
def valfilter[K: Hashable, V0, V1](predicate: Callable[[V0], TypeGuard[V1]], d: Mapping[K, V0], factory: Callable[[], MutableMapping[K, V1]]) -> MutableMapping[K, V1]: ...
@overload
def valfilter[K: Hashable, V0, V1](predicate: Callable[[V0], bool], d: Mapping[K, V0], factory: Callable[[], MutableMapping[K, V1]]) -> MutableMapping[K, V1]: ...
def valfilter[K: Hashable, V0, V1](predicate: Callable[[V0], bool] | Callable[[V0], TypeGuard[V1]] | Callable[[V0], TypeIs[V1]], d: Mapping[K, V0], factory: Callable[[], MutableMapping[K, V1]] = dict) -> MutableMapping[K, V1]:
	"""Retain only items from `d` whose values satisfy `predicate` and return a new `Mapping`.

	(AI generated docstring)

	You can use `valfilter` to select items from `d` (***d***ictionary) by their values. `valfilter`
	calls `predicate` with each value yielded by `d.values()`. `valfilter` inserts each item whose
	value causes `predicate` to return `True` into a new `MutableMapping`[1] created by `factory`.
	`valfilter` does not change `d`.

	Parameters
	----------
	predicate : Callable[[V], bool]
		`Callable` applied to each value of `d`. `valfilter` keeps each item for which `predicate`
		returns `True`.
	d : Mapping[K, V]
		Source `Mapping`[1]. `valfilter` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingFiltered : MutableMapping[K, V]
		New `MutableMapping` created by `factory` containing only items from `d` whose values
		cause `predicate` to return `True`.

	See Also
	--------
	keyfilter : Retain only items from `d` whose keys satisfy `predicate` and return a new `Mapping`.
	itemfilter : Retain only items from `d` whose key-value pairs satisfy `predicate` and return a new `Mapping`.
	valmap : Apply a `Callable` to all values of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> iseven = lambda x: x % 2 == 0
	>>> d = {1: 2, 2: 3, 3: 4, 4: 5}
	>>> valfilter(iseven, d)
	{1: 2, 3: 4}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	rv: MutableMapping[K, V1] = factory()
	for k, v in d.items():
		if predicate(v):
			# TODO Is `cast` appropriate?
			rv[k] = v  # pyright: ignore[reportArgumentType] # ty:ignore[invalid-assignment]
	return rv

@overload
def valmap[K: Hashable, V0, V1](func: Callable[[V0], V1], d: Mapping[K, V0], factory: Callable[[], dict[K, V1]] = dict) -> dict[K, V1]: ...
@overload
def valmap[K: Hashable, V0, V1](func: Callable[[V0], V1], d: Mapping[K, V0], factory: Callable[[], MutableMapping[K, V1]]) -> MutableMapping[K, V1]: ...
def valmap[K: Hashable, V0, V1](func: Callable[[V0], V1], d: Mapping[K, V0], factory: Callable[[], MutableMapping[K, V1]] = dict) -> MutableMapping[K, V1]:
	"""Apply `func` to all values of `d` and return a new `Mapping` with the transformed values.

	(AI generated docstring)

	You can use `valmap` to transform all values in `d` without changing `d`. `valmap` applies
	`func` to each value yielded by `d.values()` and builds a new `MutableMapping`[1] created
	by `factory`, preserving each original key associated with its transformed value.

	Parameters
	----------
	func : Callable[[V], W]
		`Callable` applied to each value of `d`. Each value of `d` is passed to `func`
		individually, and `func` returns the corresponding transformed value.
	d : Mapping[K, V]
		Source `Mapping`[1]. `valmap` reads all values from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, W]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingTransformed : MutableMapping[K, W]
		New `MutableMapping` created by `factory` in which each key from `d` is associated
		with the result of applying `func` to the corresponding value of `d`.

	See Also
	--------
	keymap : Apply a `Callable` to all keys of a `Mapping` and return a new `Mapping`.
	itemmap : Apply a `Callable` to all items of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> bills = {'Alice': [20, 15, 30], 'Bob': [10, 35]}
	>>> valmap(sum, bills)  # doctest: +SKIP
	{'Alice': 65, 'Bob': 45}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	rv: MutableMapping[K, V1] = factory()
	rv.update(zip(d.keys(), map(func, d.values()), strict=True))
	return rv




"""Tests for the coping mechanism module.

(AI generated docstring)

This module validates the behavior of package setting retrieval,
null-check utilities, and installation path resolution.

"""
from hunterMakesPy import PackageSettings, raiseIfNone
from hunterMakesPy.coping import getIdentifierPackagePACKAGING, getPathPackageINSTALLING
from hunterMakesPy.tests.conftest import uniformTestFailureMessage
from pathlib import Path
import pytest

@pytest.mark.parametrize(
	"returnTarget, expected",
	[
		(13, 13),
		(17, 17),
		("fibonacci", "fibonacci"),
		("prime", "prime"),
		([], []),
		({}, {}),
		(False, False),
		(0, 0),
	]
)
def testRaiseIfNoneReturnsNonNoneValues(returnTarget: object, expected: object) -> None:
	"""Verify that non-None values are returned exactly as provided.

	(AI generated docstring)

	Parameters
	----------
	returnTarget : object
		The value to pass to `raiseIfNone`.
	expected : object
		The expected return value (should be identical to `returnTarget`).

	"""
	actual: object = raiseIfNone(returnTarget)
	assert actual == expected, uniformTestFailureMessage(expected, actual, "testRaiseIfNoneReturnsNonNoneValues", returnTarget)
	assert actual is returnTarget, uniformTestFailureMessage(returnTarget, actual, "testRaiseIfNoneReturnsNonNoneValues identity check", returnTarget)

def testRaiseIfNoneRaisesValueErrorWhenGivenNone() -> None:
	"""Verify that ValueError is raised when input is None.

	(AI generated docstring)

	"""
	with pytest.raises(ValueError, match=r"A function unexpectedly returned `None`.*"):
		raiseIfNone(None)

@pytest.mark.parametrize(
	"customMessage",
	[
		"Configuration must include 'host' setting",
		"Database connection failed",
		"User input is required",
		"Network request returned empty response",
	]
)
def testRaiseIfNoneRaisesValueErrorWithCustomMessage(customMessage: str) -> None:
	"""Verify that custom error messages are used when provided.

	(AI generated docstring)

	Parameters
	----------
	customMessage : str
		The custom error message to expect.

	"""
	with pytest.raises(ValueError, match=customMessage):
		raiseIfNone(None, customMessage)

def testRaiseIfNoneWithEmptyStringMessage() -> None:
	"""Verify that empty custom message string triggers default message.

	(AI generated docstring)

	"""
	with pytest.raises(ValueError, match=r"A function unexpectedly returned `None`.*"):
		raiseIfNone(None, "")

def testRaiseIfNonePreservesTypeAnnotations() -> None:
	"""Verify that type info is preserved through the pass-through.

	(AI generated docstring)

	"""
	integerValue: int = raiseIfNone(23)
	assert isinstance(integerValue, int), uniformTestFailureMessage(int, type(integerValue), "testRaiseIfNonePreservesTypeAnnotations", integerValue)

	stringValue: str = raiseIfNone("cardinal")
	assert isinstance(stringValue, str), uniformTestFailureMessage(str, type(stringValue), "testRaiseIfNonePreservesTypeAnnotations", stringValue)

	listValue: list[int] = raiseIfNone([29, 31])
	assert isinstance(listValue, list), uniformTestFailureMessage(list, type(listValue), "testRaiseIfNonePreservesTypeAnnotations", listValue)

# Tests for PackageSettings dataclass
@pytest.mark.parametrize(
	"identifierPackageFALLBACK, expectedIdentifierPackage",
	[
		("astToolFactory", "hunterMakesPy"),  # Should read from pyproject.toml
		("nonExistentPackage", "hunterMakesPy"),  # Should read from pyproject.toml
		("customPackage", "hunterMakesPy"),  # Should read from pyproject.toml
	]
)
def testPackageSettingsWithFallbackUsesProjectToml(identifierPackageFALLBACK: str, expectedIdentifierPackage: str) -> None:
	"""Test that PackageSettings reads package name from pyproject.toml when using fallback.

	Parameters
	----------
	identifierPackageFALLBACK : str
		The fallback identifier to use if retrieval fails (or to trigger lookup).
	expectedIdentifierPackage : str
		The expected resolved package identifier.

	"""
	packageSettings: PackageSettings = PackageSettings(identifierPackageFALLBACK)
	assert packageSettings.identifierPackage == expectedIdentifierPackage, uniformTestFailureMessage(
		expectedIdentifierPackage, packageSettings.identifierPackage, "PackageSettings fallback", identifierPackageFALLBACK
	)

@pytest.mark.parametrize(
	"explicitIdentifierPackage, expectedIdentifierPackage",
	[
		("customPackageName", "customPackageName"),
		("fibonacci", "fibonacci"),
		("prime", "prime"),
		("astToolFactory", "astToolFactory"),
	]
)
def testPackageSettingsWithExplicitIdentifierPackage(explicitIdentifierPackage: str, expectedIdentifierPackage: str) -> None:
	"""Test that PackageSettings respects explicitly provided identifierPackage.

	Parameters
	----------
	explicitIdentifierPackage : str
		The explicitly provided package identifier.
	expectedIdentifierPackage : str
		The expected resolved package identifier.

	"""
	packageSettings: PackageSettings = PackageSettings(identifierPackage=explicitIdentifierPackage)
	assert packageSettings.identifierPackage == expectedIdentifierPackage, uniformTestFailureMessage(
		expectedIdentifierPackage, packageSettings.identifierPackage, "PackageSettings explicit identifierPackage", explicitIdentifierPackage
	)

@pytest.mark.parametrize(
	"explicitPathPackage, expectedPathPackage",
	[
		(Path("C:/fibonacci/path"), Path("C:/fibonacci/path")),
		(Path("C:/prime/directory"), Path("C:/prime/directory")),
		(Path("/usr/local/lib/package"), Path("/usr/local/lib/package")),
		(Path("relative/path"), Path("relative/path")),
	]
)
def testPackageSettingsWithExplicitPathPackage(explicitPathPackage: Path, expectedPathPackage: Path) -> None:
	"""Test that PackageSettings respects explicitly provided pathPackage.

	Parameters
	----------
	explicitPathPackage : Path
		The explicitly provided package path.
	expectedPathPackage : Path
		The expected resolved package path.

	"""
	packageSettings: PackageSettings = PackageSettings(pathPackage=explicitPathPackage)
	assert packageSettings.pathPackage == expectedPathPackage, uniformTestFailureMessage(
		expectedPathPackage, packageSettings.pathPackage, "PackageSettings explicit pathPackage", explicitPathPackage
	)

@pytest.mark.parametrize(
	"fileExtension, expectedFileExtension",
	[
		(".fibonacci", ".fibonacci"),
		(".prime", ".prime"),
		(".txt", ".txt"),
		(".md", ".md"),
		(".json", ".json"),
	]
)
def testPackageSettingsWithCustomFileExtension(fileExtension: str, expectedFileExtension: str) -> None:
	"""Test that PackageSettings respects custom file extensions.

	Parameters
	----------
	fileExtension : str
		The custom file extension to set.
	expectedFileExtension : str
		The expected file extension in the settings.

	"""
	packageSettings: PackageSettings = PackageSettings(fileExtension=fileExtension)
	assert packageSettings.fileExtension == expectedFileExtension, uniformTestFailureMessage(
		expectedFileExtension, packageSettings.fileExtension, "PackageSettings custom fileExtension", fileExtension
	)

def testPackageSettingsDefaultValues() -> None:
	"""Test that PackageSettings has correct default values when no arguments provided."""
	packageSettings: PackageSettings = PackageSettings()

	# Should have default file extension
	assert packageSettings.fileExtension == '.py', uniformTestFailureMessage(
		'.py', packageSettings.fileExtension, "PackageSettings default fileExtension"
	)

	# identifierPackage should be empty when no fallback provided
	assert packageSettings.identifierPackage == '', uniformTestFailureMessage(
		'', packageSettings.identifierPackage, "PackageSettings default identifierPackage"
	)

	expectedPath: Path = Path()
	assert packageSettings.pathPackage == expectedPath, uniformTestFailureMessage(
		expectedPath, packageSettings.pathPackage, "PackageSettings default pathPackage, pathPackage should remain as Path() when identifierPackage is empty"
	)

@pytest.mark.parametrize(
	"identifierPackageFALLBACK, identifierPackage, pathPackage, fileExtension",
	[
		("fallback", "custom", Path("C:/custom/path"), ".txt"),
		("fibonacci", "prime", Path("C:/fibonacci/lib"), ".md"),
		("defaultFallback", "overridePackage", Path("/usr/local/override"), ".json"),
	]
)
def testPackageSettingsAllParametersCombined(
	identifierPackageFALLBACK: str,
	identifierPackage: str,
	pathPackage: Path,
	fileExtension: str
) -> None:
	"""Test PackageSettings with all parameters provided.

	Parameters
	----------
	identifierPackageFALLBACK : str
		The fallback identifier.
	identifierPackage : str
		The explicit package identifier.
	pathPackage : Path
		The explicit package path.
	fileExtension : str
		The explicit file extension.

	"""
	packageSettings: PackageSettings = PackageSettings(
		identifierPackageFALLBACK
		, identifierPackage=identifierPackage
		, pathPackage=pathPackage
		, fileExtension=fileExtension
	)

	assert packageSettings.identifierPackage == identifierPackage, uniformTestFailureMessage(
		identifierPackage, packageSettings.identifierPackage, "PackageSettings combined identifierPackage"
	)
	assert packageSettings.pathPackage == pathPackage, uniformTestFailureMessage(
		pathPackage, packageSettings.pathPackage, "PackageSettings combined pathPackage"
	)
	assert packageSettings.fileExtension == fileExtension, uniformTestFailureMessage(
		fileExtension, packageSettings.fileExtension, "PackageSettings combined fileExtension"
	)

def testPackageSettingsFallbackIgnoredWhenExplicitIdentifierProvided() -> None:
	"""Test that fallback is ignored when explicit identifierPackage is provided."""
	packageSettings: PackageSettings = PackageSettings("shouldBeIgnored", identifierPackage="explicit")
	assert packageSettings.identifierPackage == "explicit", uniformTestFailureMessage(
		"explicit", packageSettings.identifierPackage, "PackageSettings fallback ignored"
	)

@pytest.mark.parametrize(
	"identifierPackageFALLBACK, expectedResult",
	[
		("fibonacci", "hunterMakesPy"),  # Should read from pyproject.toml
		("prime", "hunterMakesPy"),  # Should read from pyproject.toml
		("nonExistentPackage", "hunterMakesPy"),  # Should read from pyproject.toml
	]
)
def testGetIdentifierPackagePACKAGING(identifierPackageFALLBACK: str, expectedResult: str) -> None:
	"""Test that getIdentifierPackagePACKAGING reads from pyproject.toml correctly.

	Parameters
	----------
	identifierPackageFALLBACK : str
		The fallback identifier to provide.
	expectedResult : str
		The expected package identifier result.

	"""
	actual: str = getIdentifierPackagePACKAGING(identifierPackageFALLBACK)
	assert actual == expectedResult, uniformTestFailureMessage(
		expectedResult, actual, "getIdentifierPackagePACKAGING", identifierPackageFALLBACK
	)

@pytest.mark.parametrize(
	"identifierPackage",
	[
		"hunterMakesPy",  # This package exists
		"fibonacci",  # Non-existent package should fallback to cwd
		"prime",  # Non-existent package should fallback to cwd
	]
)
def testGetPathPackageINSTALLING(identifierPackage: str) -> None:
	"""Test that getPathPackageINSTALLING returns valid Path objects.

	Parameters
	----------
	identifierPackage : str
		The package identifier to look up.

	"""
	actual: Path = getPathPackageINSTALLING(identifierPackage)
	assert isinstance(actual, Path), uniformTestFailureMessage(
		Path, type(actual), "getPathPackageINSTALLING type", identifierPackage
	)
	assert actual.exists() or actual == Path.cwd(), uniformTestFailureMessage(
		"existing path or cwd", actual, "getPathPackageINSTALLING existence", identifierPackage
	)



# from .dev_skip_test import dev_skip_test
# from humpy_cytoolz import curry, identity, keyfilter, merge_with, valfilter
# from types import BuiltinFunctionType, FunctionType
# import humpy_cytoolz
# import inspect

# @curry
# def isfrommod(modname, func):
#     mod = getattr(func, '__module__', '') or ''
#     return mod.startswith(modname) or 'toolz.functoolz.curry' in str(type(func))

# @dev_skip_test
# def test_class_sigs():
#     """ Test that all ``cdef class`` extension types in ``humpy_cytoolz`` have
#         correctly embedded the function signature as done in ``humpy_toolz``.
#     """
#     import humpy_toolz
#     toolz_dict = valfilter(isfrommod('humpy_toolz'), humpy_toolz.__dict__)
#     cytoolz_dict = valfilter(isfrommod('humpy_cytoolz'), humpy_cytoolz.__dict__)
#     cytoolz_dict = valfilter(lambda x: not isinstance(x, BuiltinFunctionType), cytoolz_dict)
#     toolz_dict = keyfilter(lambda x: x in cytoolz_dict, toolz_dict)
#     cytoolz_dict = keyfilter(lambda x: x in toolz_dict, cytoolz_dict)

#     class wrap:
#         """e.g., allow `factory=<class 'dict'>` to instead be `factory=dict` in signature"""

#         def __init__(self, obj):
#             self.obj = obj

#         def __repr__(self):
#             return getattr(self.obj, '__name__', repr(self.obj))
#     d = merge_with(identity, toolz_dict, cytoolz_dict)
#     for key, (toolz_func, cytoolz_func) in d.items():
#         if key in {'__getattr__'}:
#             continue
#         if isinstance(toolz_func, FunctionType):
#             toolz_spec = inspect.signature(toolz_func)
#         elif isinstance(toolz_func, humpy_toolz.curry):
#             toolz_spec = inspect.signature(toolz_func.func)
#         else:
#             toolz_spec = inspect.signature(toolz_func.__init__)
#         toolz_spec = toolz_spec.replace(parameters=[v.replace(default=wrap(v.default)) if v.default is not inspect._empty else v for v in toolz_spec.parameters.values()])
#         doc = cytoolz_func.__doc__
#         doc_alt = doc.replace('Py_ssize_t ', '').replace("=u'", "='")
#         toolz_sig = toolz_func.__name__ + str(toolz_spec)
#         if not (toolz_sig in doc or toolz_sig in doc_alt):
#             message = 'cytoolz.%s does not have correct function signature.\n\nExpected: %s\n\nDocstring in cytoolz is:\n%s' % (key, toolz_sig, cytoolz_func.__doc__)
#             assert False, message
# skip_sigs = ['identity']
# aliases = {'comp': 'compose'}

# @dev_skip_test
# def test_sig_at_beginning():
#     """ Test that the function signature is at the beginning of the docstring
#         and is followed by exactly one blank line.
#     """
#     cytoolz_dict = valfilter(isfrommod('humpy_cytoolz'), humpy_cytoolz.__dict__)
#     cytoolz_dict = keyfilter(lambda x: x not in skip_sigs, cytoolz_dict)
#     for key, val in cytoolz_dict.items():
#         if key in {'__getattr__'}:
#             continue
#         doclines = val.__doc__.splitlines()
#         assert len(doclines) > 2, f'cytoolz.{key} docstring too short:\n\n{val.__doc__}'
#         sig = '%s(' % aliases.get(key, key)
#         assert sig in doclines[0], 'cytoolz.%s docstring missing signature at beginning:\n\n%s' % (key, val.__doc__)
#         assert not doclines[1], 'cytoolz.%s docstring missing blank line after signature:\n\n%s' % (key, val.__doc__)
#         assert doclines[2], 'cytoolz.%s docstring too many blank lines after signature:\n\n%s' % (key, val.__doc__)



import humpy_toolz

__all__ = ['merge_with', 'merge']

@humpy_toolz.curry
def merge_with(func, d, *dicts, **kwargs):
    return humpy_toolz.merge_with(func, d, *dicts, **kwargs)

@humpy_toolz.curry
def merge(d, *dicts, **kwargs):
    return humpy_toolz.merge(d, *dicts, **kwargs)
merge_with.__doc__ = humpy_toolz.merge_with.__doc__
merge.__doc__ = humpy_toolz.merge.__doc__



from collections.abc import Callable, Iterable, Iterator
from functools import partial
from humpy_toolz.itertoolz import (
	accumulate, concat, concatv, cons, count, diff, drop, first, frequencies, get, getter, groupby, interleave, interpose, isdistinct,
	isiterable, iterate, join, last, mapcat, merge_sorted, nth, partition, partition_all, peek, peekn, pluck, random_sample, reduceby, remove,
	rest, second, sliding_window, tail, take, take_nth, topk, unique)
from humpy_toolz.utils import raises
from itertools import starmap
from operator import add, mul
from pickle import dumps, loads
from random import Random
from typing import Any, NoReturn
import itertools

no_default2 = loads(dumps('__no__default__'))

def identity(x: object) -> object:
    return x

def iseven(x: int) -> bool:
    return x % 2 == 0

def isodd(x: int) -> bool:
    return x % 2 == 1

def inc(x: int) -> int:
    return x + 1

def double(x: int) -> int:
    return 2 * x

def test_remove() -> None:
    r: Iterable[int] = remove(iseven, range(5))
    assert type(r) is not list
    assert list(r) == list(filter(isodd, range(5)))

def test_groupby() -> None:
    assert groupby(iseven, [1, 2, 3, 4]) == {True: [2, 4], False: [1, 3]}

def test_groupby_non_callable() -> None:
    assert groupby(0, [(1, 2), (1, 3), (2, 2), (2, 4)]) == {1: [(1, 2), (1, 3)], 2: [(2, 2), (2, 4)]}
    assert groupby([0], [(1, 2), (1, 3), (2, 2), (2, 4)]) == {(1,): [(1, 2), (1, 3)], (2,): [(2, 2), (2, 4)]}
    assert groupby([0, 0], [(1, 2), (1, 3), (2, 2), (2, 4)]) == {(1, 1): [(1, 2), (1, 3)], (2, 2): [(2, 2), (2, 4)]}

def test_merge_sorted() -> None:
    assert list(merge_sorted([1, 2, 3], [1, 2, 3])) == [1, 1, 2, 2, 3, 3]
    assert list(merge_sorted([1, 3, 5], [2, 4, 6])) == [1, 2, 3, 4, 5, 6]
    assert list(merge_sorted([1], [2, 4], [3], [])) == [1, 2, 3, 4]
    assert list(merge_sorted([5, 3, 1], [6, 4, 3], [], key=lambda x: -x)) == [6, 5, 4, 3, 3, 1]
    assert list(merge_sorted([2, 1, 3], [1, 2, 3], key=lambda x: x // 3)) == [2, 1, 1, 2, 3, 3]
    assert list(merge_sorted([2, 3], [1, 3], key=lambda x: x // 3)) == [2, 1, 3, 3]
    assert ''.join(merge_sorted('abc', 'abc', 'abc')) == 'aaabbbccc'
    assert ''.join(merge_sorted('abc', 'abc', 'abc', key=ord)) == 'aaabbbccc'
    assert ''.join(merge_sorted('cba', 'cba', 'cba', key=lambda x: -ord(x))) == 'cccbbbaaa'
    assert list(merge_sorted([1], [2, 3, 4], key=identity)) == [1, 2, 3, 4]
    data: list[list[tuple[int, int]]] = [[(1, 2), (0, 4), (3, 6)], [(5, 3), (6, 5), (8, 8)], [(9, 1), (9, 8), (9, 9)]]
    assert list(merge_sorted(*data, key=lambda x: x[1])) == [(9, 1), (1, 2), (5, 3), (0, 4), (6, 5), (3, 6), (8, 8), (9, 8), (9, 9)]
    assert list(merge_sorted()) == []
    assert list(merge_sorted([1, 2, 3])) == [1, 2, 3]
    assert list(merge_sorted([1, 4, 5], [2, 3])) == [1, 2, 3, 4, 5]
    assert list(merge_sorted([1, 4, 5], [2, 3], key=identity)) == [1, 2, 3, 4, 5]
    assert list(merge_sorted([1, 5], [2], [4, 7], [3, 6], key=identity)) == [1, 2, 3, 4, 5, 6, 7]

def test_interleave() -> None:
    assert ''.join(interleave(('ABC', '123'))) == 'A1B2C3'
    assert ''.join(interleave(('ABC', '1'))) == 'A1BC'

def test_unique() -> None:
    assert tuple(unique((1, 2, 3))) == (1, 2, 3)
    assert tuple(unique((1, 2, 1, 3))) == (1, 2, 3)
    assert tuple(unique((1, 2, 3), key=iseven)) == (1, 2)

def test_isiterable() -> None:

    class IterIterable:

        def __iter__(self) -> Iterator[str]:
            return iter(['a', 'b', 'c'])

    class GetItemIterable:

        def __getitem__(self, item: int) -> str:
            return ['a', 'b', 'c'][item]

    class NotIterable:
        __iter__ = None

    class NotIterableEvenWithGetItem:
        __iter__ = None

        def __getitem__(self, item: int) -> str:
            return ['a', 'b', 'c'][item]
    assert isiterable([1, 2, 3]) is True
    assert isiterable('abc') is True
    assert isiterable(IterIterable()) is True
    assert isiterable(GetItemIterable()) is True
    assert isiterable(5) is False
    assert isiterable(NotIterable()) is False
    assert isiterable(NotIterableEvenWithGetItem()) is False

def test_isdistinct() -> None:
    assert isdistinct([1, 2, 3]) is True
    assert isdistinct([1, 2, 1]) is False
    assert isdistinct('Hello') is False
    assert isdistinct('World') is True
    assert isdistinct(iter([1, 2, 3])) is True
    assert isdistinct(iter([1, 2, 1])) is False

def test_nth() -> None:
    assert nth(2, 'ABCDE') == 'C'
    assert nth(2, iter('ABCDE')) == 'C'
    assert nth(1, (3, 2, 1)) == 2
    assert nth(0, {'foo': 'bar'}) == 'foo'
    assert raises(StopIteration, lambda: nth(10, {10: 'foo'}))
    assert nth(-2, 'ABCDE') == 'D'
    assert raises(ValueError, lambda: nth(-2, iter('ABCDE')))

def test_first() -> None:
    assert first('ABCDE') == 'A'
    assert first((3, 2, 1)) == 3
    assert isinstance(first({0: 'zero', 1: 'one'}), int)

def test_second() -> None:
    assert second('ABCDE') == 'B'
    assert second((3, 2, 1)) == 2
    assert isinstance(second({0: 'zero', 1: 'one'}), int)

def test_last() -> None:
    assert last('ABCDE') == 'E'
    assert last((3, 2, 1)) == 1
    assert isinstance(last({0: 'zero', 1: 'one'}), int)

def test_rest() -> None:
    assert list(rest('ABCDE')) == list('BCDE')
    assert list(rest((3, 2, 1))) == list((2, 1))

def test_take() -> None:
    assert list(take(3, 'ABCDE')) == list('ABC')
    assert list(take(2, (3, 2, 1))) == list((3, 2))

def test_tail() -> None:
    assert list(tail(3, 'ABCDE')) == list('CDE')
    assert list(tail(3, iter('ABCDE'))) == list('CDE')
    assert list(tail(2, (3, 2, 1))) == list((2, 1))

def test_drop() -> None:
    assert list(drop(3, 'ABCDE')) == list('DE')
    assert list(drop(1, (3, 2, 1))) == list((2, 1))

def test_take_nth() -> None:
    assert list(take_nth(2, 'ABCDE')) == list('ACE')

def test_get() -> None:
    assert get(1, 'ABCDE') == 'B'
    assert list(get([1, 3], 'ABCDE')) == list('BD')
    assert get('a', {'a': 1, 'b': 2, 'c': 3}) == 1
    assert get(['a', 'b'], {'a': 1, 'b': 2, 'c': 3}) == (1, 2)
    assert get('foo', {}, default='bar') == 'bar'
    assert get({}, [1, 2, 3], default='bar') == 'bar'
    assert get([0, 2], 'AB', 'C') == ('A', 'C')
    assert get([0], 'AB') == ('A',)
    assert get([], 'AB') == ()
    assert raises(IndexError, lambda: get(10, 'ABC'))
    assert raises(KeyError, lambda: get(10, {'a': 1}))
    assert raises(TypeError, lambda: get({}, [1, 2, 3]))
    assert raises(TypeError, lambda: get([1, 2, 3], 1, None))
    assert raises(KeyError, lambda: get('foo', {}, default=no_default2))

def test_mapcat() -> None:
    assert list(mapcat(identity, [[1, 2, 3], [4, 5, 6]])) == [1, 2, 3, 4, 5, 6]
    assert list(mapcat(reversed, [[3, 2, 1, 0], [6, 5, 4], [9, 8, 7]])) == list(range(10))
    inc = lambda i: i + 1
    assert [4, 5, 6, 7, 8, 9] == list(mapcat(partial(map, inc), [[3, 4, 5], [6, 7, 8]]))

def test_cons() -> None:
    assert list(cons(1, [2, 3])) == [1, 2, 3]

def test_concat() -> None:
    assert list(concat([[], [], []])) == []
    assert list(take(5, concat([['a', 'b'], range(1000000000)]))) == ['a', 'b', 0, 1, 2]

def test_concatv() -> None:
    assert list(concatv([], [], [])) == []
    assert list(take(5, concatv(['a', 'b'], range(1000000000)))) == ['a', 'b', 0, 1, 2]

def test_interpose() -> None:
    assert 'a' == first(rest(interpose('a', range(1000000000))))
    assert 'tXaXrXzXaXn' == ''.join(interpose('X', 'tarzan'))
    assert list(interpose(0, itertools.repeat(1, 4))) == [1, 0, 1, 0, 1, 0, 1]
    assert list(interpose('.', ['a', 'b', 'c'])) == ['a', '.', 'b', '.', 'c']

def test_frequencies() -> None:
    assert frequencies(['cat', 'pig', 'cat', 'eel', 'pig', 'dog', 'dog', 'dog']) == {'cat': 2, 'eel': 1, 'pig': 2, 'dog': 3}
    assert frequencies([]) == {}
    assert frequencies('onomatopoeia') == {'a': 2, 'e': 1, 'i': 1, 'm': 1, 'o': 4, 'n': 1, 'p': 1, 't': 1}

def test_reduceby() -> None:
    data: list[int] = [1, 2, 3, 4, 5]
    iseven = lambda x: x % 2 == 0
    assert reduceby(iseven, add, data, 0) == {False: 9, True: 6}
    assert reduceby(iseven, mul, data, 1) == {False: 15, True: 8}
    projects = [{'name': 'build roads', 'state': 'CA', 'cost': 1000000}, {'name': 'fight crime', 'state': 'IL', 'cost': 100000}, {'name': 'help farmers', 'state': 'IL', 'cost': 2000000}, {'name': 'help farmers', 'state': 'CA', 'cost': 200000}]
    assert reduceby(lambda x: x['state'], lambda acc, x: acc + x['cost'], projects, 0) == {'CA': 1200000, 'IL': 2100000}
    assert reduceby('state', lambda acc, x: acc + x['cost'], projects, 0) == {'CA': 1200000, 'IL': 2100000}

def test_reduce_by_init() -> None:
    assert reduceby(iseven, add, [1, 2, 3, 4]) == {True: 2 + 4, False: 1 + 3}
    assert reduceby(iseven, add, [1, 2, 3, 4], no_default2) == {True: 2 + 4, False: 1 + 3}

def test_reduce_by_callable_default() -> None:

    def set_add(s: set[int], i: int) -> set[int]:
        s.add(i)
        return s
    assert reduceby(iseven, set_add, [1, 2, 3, 4, 1, 2], set) == {True: {2, 4}, False: {1, 3}}

def test_iterate() -> None:
    assert list(itertools.islice(iterate(inc, 0), 0, 5)) == [0, 1, 2, 3, 4]
    assert list(take(4, iterate(double, 1))) == [1, 2, 4, 8]

def test_accumulate() -> None:
    assert list(accumulate(add, [1, 2, 3, 4, 5])) == [1, 3, 6, 10, 15]
    assert list(accumulate(mul, [1, 2, 3, 4, 5])) == [1, 2, 6, 24, 120]
    assert list(accumulate(add, [1, 2, 3, 4, 5], -1)) == [-1, 0, 2, 5, 9, 14]

    def binop(a: object, b: object) -> NoReturn:
        raise AssertionError('binop should not be called')
    start: object = object()
    assert list(accumulate(binop, [], start)) == [start]
    assert list(accumulate(binop, [])) == []
    assert list(accumulate(add, [1, 2, 3], no_default2)) == [1, 3, 6]

def test_accumulate_works_on_consumable_iterables() -> None:
    assert list(accumulate(add, iter((1, 2, 3)))) == [1, 3, 6]

def test_sliding_window() -> None:
    assert list(sliding_window(2, [1, 2, 3, 4])) == [(1, 2), (2, 3), (3, 4)]
    assert list(sliding_window(3, [1, 2, 3, 4])) == [(1, 2, 3), (2, 3, 4)]

def test_sliding_window_of_short_iterator() -> None:
    assert list(sliding_window(3, [1, 2])) == []
    assert list(sliding_window(7, [1, 2])) == []

def test_partition() -> None:
    assert list(partition(2, [1, 2, 3, 4])) == [(1, 2), (3, 4)]
    assert list(partition(3, range(7))) == [(0, 1, 2), (3, 4, 5)]
    assert list(partition(3, range(4), pad=-1)) == [(0, 1, 2), (3, -1, -1)]
    assert list(partition(2, [])) == []

def test_partition_all() -> None:
    assert list(partition_all(2, [1, 2, 3, 4])) == [(1, 2), (3, 4)]
    assert list(partition_all(3, range(5))) == [(0, 1, 2), (3, 4)]
    assert list(partition_all(2, [])) == []

    class NoCompare:

        def __eq__(self, other: object) -> bool:
            if self.__class__ == other.__class__:
                return True
            raise ValueError
    obj: NoCompare = NoCompare()
    result: list[tuple[NoCompare, ...]] = [(obj, obj, obj, obj), (obj, obj, obj)]
    assert list(partition_all(4, [obj] * 7)) == result
    assert list(partition_all(4, iter([obj] * 7))) == result

    class ListWithBadLength(list):

        def __init__(self, contents: list[int], off_by: int = 1) -> None:
            self.off_by = off_by
            super().__init__(contents)

        def __len__(self) -> int:
            return super().__len__() + self.off_by
    too_long_list: ListWithBadLength = ListWithBadLength([1, 2], off_by=+1)
    assert raises(LookupError, lambda: list(partition_all(5, too_long_list)))
    too_short_list: ListWithBadLength = ListWithBadLength([1, 2], off_by=-1)
    assert raises(LookupError, lambda: list(partition_all(5, too_short_list)))

def test_count() -> None:
    assert count((1, 2, 3)) == 3
    assert count([]) == 0
    assert count(iter((1, 2, 3, 4))) == 4
    assert count('hello') == 5
    assert count(iter('hello')) == 5

def test_pluck() -> None:
    assert list(pluck(0, [[0, 1], [2, 3], [4, 5]])) == [0, 2, 4]
    assert list(pluck([0, 1], [[0, 1, 2], [3, 4, 5]])) == [(0, 1), (3, 4)]
    assert list(pluck(1, [[0], [0, 1]], None)) == [None, 1]
    data = [{'id': 1, 'name': 'cheese'}, {'id': 2, 'name': 'pies', 'price': 1}]
    assert list(pluck('id', data)) == [1, 2]
    assert list(pluck('price', data, 0)) == [0, 1]
    assert list(pluck(['id', 'name'], data)) == [(1, 'cheese'), (2, 'pies')]
    assert list(pluck(['name'], data)) == [('cheese',), ('pies',)]
    assert list(pluck(['price', 'other'], data, 0)) == [(0, 0), (1, 0)]
    assert raises(IndexError, lambda: list(pluck(1, [[0]])))
    assert raises(KeyError, lambda: list(pluck('name', [{'id': 1}])))
    assert list(pluck(0, [[0, 1], [2, 3], [4, 5]], no_default2)) == [0, 2, 4]
    assert raises(IndexError, lambda: list(pluck(1, [[0]], no_default2)))

def test_join() -> None:
    names: list[tuple[int, str]] = [(1, 'one'), (2, 'two'), (3, 'three')]
    fruit: list[tuple[str, int]] = [('apple', 1), ('orange', 1), ('banana', 2), ('coconut', 2)]

    def addpair(pair: tuple[Any, Any]) -> Any:
        return pair[0] + pair[1]
    result: set[tuple[int, str, str, int]] = set(starmap(add, join(first, names, second, fruit)))
    expected: set[tuple[int, str, str, int]] = {(1, 'one', 'apple', 1), (1, 'one', 'orange', 1), (2, 'two', 'banana', 2), (2, 'two', 'coconut', 2)}
    assert result == expected
    result = set(starmap(add, join(first, names, second, fruit, left_default=no_default2, right_default=no_default2)))
    assert result == expected

def test_getter() -> None:
    assert getter(0)('Alice') == 'A'
    assert getter([0])('Alice') == ('A',)
    assert getter([])('Alice') == ()

def test_key_as_getter() -> None:
    squares: list[tuple[int, int]] = [(i, i ** 2) for i in range(5)]
    pows: list[tuple[int, int, int]] = [(i, i ** 2, i ** 3) for i in range(5)]
    assert set(join(0, squares, 0, pows)) == set(join(lambda x: x[0], squares, lambda x: x[0], pows))
    get = lambda x: (x[0], x[1])
    assert set(join([0, 1], squares, [0, 1], pows)) == set(join(get, squares, get, pows))
    get = lambda x: (x[0],)
    assert set(join([0], squares, [0], pows)) == set(join(get, squares, get, pows))

def test_join_double_repeats() -> None:
    names: list[tuple[int, str]] = [(1, 'one'), (2, 'two'), (3, 'three'), (1, 'uno'), (2, 'dos')]
    fruit: list[tuple[str, int]] = [('apple', 1), ('orange', 1), ('banana', 2), ('coconut', 2)]
    result: set[tuple[int, str, str, int]] = set(starmap(add, join(first, names, second, fruit)))
    expected: set[tuple[int, str, str, int]] = {(1, 'one', 'apple', 1), (1, 'one', 'orange', 1), (2, 'two', 'banana', 2), (2, 'two', 'coconut', 2), (1, 'uno', 'apple', 1), (1, 'uno', 'orange', 1), (2, 'dos', 'banana', 2), (2, 'dos', 'coconut', 2)}
    assert result == expected

def test_join_missing_element() -> None:
    names: list[tuple[int, str]] = [(1, 'one'), (2, 'two'), (3, 'three')]
    fruit: list[tuple[str, int]] = [('apple', 5), ('orange', 1)]
    result: set[tuple[int, str, str, int]] = set(starmap(add, join(first, names, second, fruit)))
    expected: set[tuple[int, str, str, int]] = {(1, 'one', 'orange', 1)}
    assert result == expected

def test_left_outer_join() -> None:
    result: set[tuple[int | None, int | None]] = set(join(identity, [1, 2], identity, [2, 3], left_default=None))
    expected: set[tuple[int | None, int | None]] = {(2, 2), (None, 3)}
    assert result == expected

def test_right_outer_join() -> None:
    result: set[tuple[int | None, int | None]] = set(join(identity, [1, 2], identity, [2, 3], right_default=None))
    expected: set[tuple[int | None, int | None]] = {(2, 2), (1, None)}
    assert result == expected

def test_outer_join() -> None:
    result: set[tuple[int | None, int | None]] = set(join(identity, [1, 2], identity, [2, 3], left_default=None, right_default=None))
    expected: set[tuple[int | None, int | None]] = {(2, 2), (1, None), (None, 3)}
    assert result == expected

def test_diff() -> None:
    assert raises(TypeError, lambda: list(diff()))
    assert raises(TypeError, lambda: list(diff([1, 2])))
    assert raises(TypeError, lambda: list(diff([1, 2], 3)))
    assert list(diff([1, 2], (1, 2), iter([1, 2]))) == []
    assert list(diff([1, 2, 3], (1, 10, 3), iter([1, 2, 10]))) == [(2, 10, 2), (3, 3, 10)]
    assert list(diff([1, 2], [10])) == [(1, 10)]
    assert list(diff([1, 2], [10], default=None)) == [(1, 10), (2, None)]
    assert raises(TypeError, lambda: list(diff([])))
    assert raises(TypeError, lambda: list(diff([[]])))
    assert raises(TypeError, lambda: list(diff([[1, 2]])))
    assert raises(TypeError, lambda: list(diff([[1, 2], 3])))
    assert list(diff([(1, 2), (1, 3)])) == [(2, 3)]
    data1: list[dict[str, int | str]] = [{'cost': 1, 'currency': 'dollar'}, {'cost': 2, 'currency': 'dollar'}]
    data2: list[dict[str, int | str]] = [{'cost': 100, 'currency': 'yen'}, {'cost': 300, 'currency': 'yen'}]
    conversions: dict[str, float] = {'dollar': 1, 'yen': 0.01}

    def indollars(item: dict[str, int | str]) -> int | float:
        return conversions[item['currency']] * item['cost']
    list(diff(data1, data2, key=indollars)) == [({'cost': 2, 'currency': 'dollar'}, {'cost': 300, 'currency': 'yen'})]

def test_topk() -> None:
    assert topk(2, [4, 1, 5, 2]) == (5, 4)
    assert topk(2, [4, 1, 5, 2], key=lambda x: -x) == (1, 2)
    assert topk(2, iter([5, 1, 4, 2]), key=lambda x: -x) == (1, 2)
    assert topk(2, [{'a': 1, 'b': 10}, {'a': 2, 'b': 9}, {'a': 10, 'b': 1}, {'a': 9, 'b': 2}], key='a') == ({'a': 10, 'b': 1}, {'a': 9, 'b': 2})
    assert topk(2, [{'a': 1, 'b': 10}, {'a': 2, 'b': 9}, {'a': 10, 'b': 1}, {'a': 9, 'b': 2}], key='b') == ({'a': 1, 'b': 10}, {'a': 2, 'b': 9})
    assert topk(2, [(0, 4), (1, 3), (2, 2), (3, 1), (4, 0)], 0) == ((4, 0), (3, 1))

def test_topk_is_stable() -> None:
    assert topk(4, [5, 9, 2, 1, 5, 3], key=lambda x: 1) == (5, 9, 2, 1)

def test_peek() -> None:
    alist: list[str] = ['Alice', 'Bob', 'Carol']
    element, blist = peek(alist)
    assert element == alist[0]
    assert list(blist) == alist
    assert raises(StopIteration, lambda: peek([]))

def test_peekn() -> None:
    alist: tuple[str, str, str] = ('Alice', 'Bob', 'Carol')
    elements, blist = peekn(2, alist)
    assert elements == alist[:2]
    assert tuple(blist) == alist
    elements, blist = peekn(len(alist) * 4, alist)
    assert elements == alist
    assert tuple(blist) == alist

def test_random_sample() -> None:
    alist: list[int] = list(range(100))
    assert list(random_sample(prob=1, seq=alist, random_state=2016)) == alist
    mk_rsample: Callable[..., list[int]] = lambda rs=1: list(random_sample(prob=0.1, seq=alist, random_state=rs))
    rsample1: list[int] = mk_rsample()
    assert rsample1 == mk_rsample()
    rsample2: list[int] = mk_rsample(1984)
    randobj: Random = Random(1984)
    assert rsample2 == mk_rsample(randobj)
    assert rsample1 != rsample2
    assert mk_rsample(hash(object)) == mk_rsample(hash(object))
    assert mk_rsample(hash(object)) != mk_rsample(hash(object()))
    assert mk_rsample(b'a') == mk_rsample('a')
    assert raises(TypeError, lambda: mk_rsample([]))



# Builtins - re-export to make them available in cytoolz namespace
from .dicttoolz import (
	assoc as assoc, assoc_in as assoc_in, dissoc as dissoc, get_in as get_in, itemfilter as itemfilter, itemmap as itemmap,
	keyfilter as keyfilter, keymap as keymap, merge as merge, merge_with as merge_with, update_in as update_in, valfilter as valfilter,
	valmap as valmap)
from .functoolz import (
	apply as apply, complement as complement, compose as compose, compose_left as compose_left, curry as curry, do as do, excepts as excepts,
	flip as flip, identity as identity, juxt as juxt, memoize as memoize, pipe as pipe, thread_first as thread_first,
	thread_last as thread_last)
from .itertoolz import (
	accumulate as accumulate, concat as concat, concatv as concatv, cons as cons, count as count, diff as diff, drop as drop, first as first,
	frequencies as frequencies, get as get, groupby as groupby, interleave as interleave, interpose as interpose, isdistinct as isdistinct,
	isiterable as isiterable, iterate as iterate, join as join, last as last, mapcat as mapcat, merge_sorted as merge_sorted, nth as nth,
	partition as partition, partition_all as partition_all, peek as peek, peekn as peekn, pluck as pluck, random_sample as random_sample,
	reduceby as reduceby, remove as remove, second as second, sliding_window as sliding_window, tail as tail, take as take,
	take_nth as take_nth, topk as topk, unique as unique)
from .recipes import countby as countby, partitionby as partitionby
from builtins import filter as filter, map as map, sorted as sorted  # noqa: A004
from functools import partial as partial, reduce as reduce

# Aliases
comp = compose

__version__: str
